; IRMP (Keyboard FDC3402)
; (C) Mario Latzig 2016
;
; on ATmega168PA
;
; PB0: LED1   (LED Block: Key Press)
; PB1: LED2   (LED Block: Caps Lock)
; PB2: CE     (SPI) 
; PB3: MOSI   (SPI)
; PB4: MISO   (SPI)
; PB5: SCK    (SPI)
;
; PC0: (unused)
; PC1: (unused)
; PC2: (unused)
; PC3: (unused)
; PC4: ONOFF  (FDC 3402 Keyboard "ON/OFF" Pulse) 
; PC5: KEYDAT (FDC 3402 Keyboard IRMP Data)  
;
; PD0: RXD    (UART Debug)
; PD1: TXD    (UART Debug)
; PD2: (unused)
; PD3: (unused)
; PD4: (unused)
; PD5: (unused)      
; PD6: (unused)    
; PD7: (unused)

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************

.equ FC = 20000000							; Clock = 20 MHz

; ### TIMER (Tick per 50us)
.equ CLK_PRELOAD = 125-1					; Preload: 125
.equ CLK_PRESCALER = (1<<CS11)				; Prescaler: 8

; ### PINS
.equ PINB_LED1 = (1<<0)						; PB0: LED1   (LED Block) 
.equ PINB_LED2 = (1<<1)						; PB1: LED2   (LED Block) 
.equ PINB_CSN = (1<<2)						; PB2: CSN    (SPI) 
.equ PINB_MOSI = (1<<3)						; PB3: MOSI   (SPI)
.equ PINB_MISO = (1<<4)						; PB4: MISO   (SPI)
.equ PINB_SCK = (1<<5)						; PB5: SCK    (SPI)

.equ PINC_ONOFF = (1<<4)					; ONOFF  (FDC 3402 Keyboard "ON/OFF" Pulse) 
.equ PINC_KEYDAT = (1<<5)					; PC5: KEYDAT (FDC 3402 Keyboard IRMP Data)  

; === IRMP ====================================================================
.equ IRMP_BITS_NUM = 40							; Number of Bits per Message
.equ IRMP_BUF_SIZE = IRMP_BITS_NUM/8			; Number of Bytes for Message Buffer

.equ IRMP_ADDR = 0x3F							; Address of FDC 3402 Keyboard
.equ IRMP_BUF_INDEX_ADDR = 0					; Message Byte #0: Address 
.equ IRMP_BUF_INDEX_RESERVED = 1				; Message Byte #1: (always Zero) 
.equ IRMP_BUF_INDEX_FLAG = 2					; Message Byte #2: Press/Release Bits (non-zero = Release)
.equ IRMP_BUF_INDEX_CMD = 3						; Message Byte #3: Command
.equ IRMP_BUF_INDEX_CMD_INV = 4					; Message Byte #4: Command (Inverted)

.equ IRMP_CMD_CAPSLOCK = 0x1E					; Command: Caps Lock
.equ IRMP_CMD_SHIFT_LEFT = 0x2C					; Command: Shift Left
.equ IRMP_CMD_SHIFT_RIGHT = 0x39				; Command: Shift Right
.equ IRMP_CMD_CTRL = 0x3A						; Command: Control
.equ IRMP_CMD_ALT = 0x3C						; Command: Alt
.equ IRMP_CMD_ALTGR = 0x3E						; Command: AltGr

.equ IRMP_KEY_CODE_UE_LO = 0x80					; Key Code: "Umlaut U" (lowercase)
.equ IRMP_KEY_CODE_OE_LO = 0x81					; Key Code: "Umalut O" (lowercase)
.equ IRMP_KEY_CODE_AE_LO = 0x82					; Key Code: "Umalut A" (lowercase)
.equ IRMP_KEY_CODE_SZ = 0x83					; Key Code: "Umlaut SZ"
.equ IRMP_KEY_CODE_CURSOR_UP = 0x84				; Key Code: Cursor Up
.equ IRMP_KEY_CODE_CURSOR_DOWN = 0x85			; Key Code: Cursor Down
.equ IRMP_KEY_CODE_CURSOR_LEFT = 0x86			; Key Code: Cursor Left
.equ IRMP_KEY_CODE_CURSOR_RIGHT = 0x87			; Key Code: Cursor Right
.equ IRMP_KEY_CODE_UE_UP = 0x88					; Key Code: "Umlaut U" (uppercase)
.equ IRMP_KEY_CODE_OE_UP = 0x89					; Key Code: "Umalut O" (uppercase)
.equ IRMP_KEY_CODE_AE_UP = 0x8A					; Key Code: "Umalut A" (uppercase)
.equ IRMP_KEY_CODE_INSERT = 0x8B				; Key Code: Insert
.equ IRMP_KEY_CODE_PAGE_UP = 0x8C				; Key Code: Page Up
.equ IRMP_KEY_CODE_PAGE_DOWN = 0x8D				; Key Code: Page Down
.equ IRMP_KEY_CODE_MOUSE_LEFT = 0x8E			; Key Code: Mouse Left
.equ IRMP_KEY_CODE_MOUSE_RIGHT = 0x8F			; Key Code: Mouse Right

.equ IRMP_KEY_CODE_MENU = 0x90					; Key Code: Menu
.equ IRMP_KEY_CODE_BACK = 0x91					; Key Code: Back
.equ IRMP_KEY_CODE_FORWARD = 0x92				; Key Code: Forward
.equ IRMP_KEY_CODE_ADDRESS = 0x93				; Key Code: Address
.equ IRMP_KEY_CODE_WINDOWS = 0x94				; Key Code: Window
.equ IRMP_KEY_CODE_FIRST_PAGE = 0x95			; Key Code: First Page
.equ IRMP_KEY_CODE_STOP = 0x96					; Key Code: Stop
.equ IRMP_KEY_CODE_MAIL = 0x97					; Key Code: Back
.equ IRMP_KEY_CODE_FAVOURITES = 0x98			; Key Code: Mail
.equ IRMP_KEY_CODE_NEW_PAGE = 0x99				; Key Code: Favourites
.equ IRMP_KEY_CODE_SETUP = 0x9A					; Key Code: Setup
.equ IRMP_KEY_CODE_FONT = 0x9B					; Key Code: Font
.equ IRMP_KEY_CODE_PRINT = 0x9C					; Key Code: Print
.equ IRMP_KEY_CODE_ONOFF = 0x9D					; Key Code: OnOff
.equ IRMP_KEY_CODE_HOME = 0x9E					; Key Code: Pos1
.equ IRMP_KEY_CODE_END = 0x9F					; Key Code: End

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

; === Clock ===================================================================
CLK_TICK:			.byte 1					; Flag for Clock Tick

; === IRMP ====================================================================
IRMP_PULSE_TICK:	.byte 1					; Flag for Pulse Tick
IRMP_PULSE_COUNT:	.byte 1					; Counter for Pulse
IRMP_PAUSE_COUNT:	.byte 1					; Counter for Pause
IRMP_PULSE_DONE:	.byte 1					; Backup for Pulse Counter (Analyze)
IRMP_PAUSE_DONE:	.byte 1					; Backup for Pause Counter (Analyze)
IRMP_BIT_COUNT:		.byte 1					; Counter for Bits
IRMP_BUF:			.byte IRMP_BUF_SIZE		; Buffer for Bits
IRMP_SHIFT:			.byte 1					; SHIFT Flag
IRMP_CTRL:			.byte 1					; CTRL Flag
IRMP_ALT:			.byte 1					; ALT Flag
IRMP_ALTGR:			.byte 1					; ALTGR Flag
IRMP_KEY:			.byte 1					; Key Code
IRMP_KEY_PRESSED:	.byte 1					; Key Code (Last Pressed)

.cseg
; *****************************************************************************
; *** Vectors                                                               ***
; *****************************************************************************

.org 0x0000
		jmp init
.org OC1Aaddr								; Interrupt "Output Compare1A" 
		jmp clk_isr_timer1                   
.org SPIaddr								; Interrupt "SPI Complete"
		jmp spi_isr_data                   
.org PCI1addr								; Interrupt "Pin Change #1" (PC5/PCINT13)
		jmp irmp_isr_pulse
.org INT_VECTORS_SIZE

; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

IRMP_CODES:									; ASCII Codes (lowercase)
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x4C,0x10,0   ,0   ,0   ,0x2B,0   ,0   	; $00-$0F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x6E,0   ,0   ,0   ,0   	; $10-$1F
	.db 0x3D,0   ,0   ,0x2A,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x1C,0x35,0x37,0x36,0   	; $20-$2F
	.db 0x0B,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0   ,0   ,0x2D,0   ,0   ,0   	; $30-$3F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $40-$4F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x01,0   	; $50-$5F
	.db 0x0D,0x1F,0x32,0x30,0x21,0x13,0x22,0x23,0x24,0x18,0x25,0x26,0x27,0x34,0x33,0x19	; $60-$6F
	.db 0x1A,0x11,0x14,0x20,0x15,0x17,0x31,0x12,0x2F,0x2E,0x16,0   ,0   ,0   ,0   ,0x0F ; $70-$7F

	.db 0x1B,0x28,0x29,0x0C,0x53,0x54,0x4F,0x59,0   ,0   ,0   ,0x4B,0x55,0x56,0   ,0   	; $80-$8F
	.db 0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x7B,0x7C,0x7E,0x50,0x51	; $90-$9F

IRMP_CODES_SHIFT:							; ASCII Codes (uppercase)
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x4C,0x10,0   ,0   ,0   ,0x2B,0   ,0   	; $00-$0F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x6E,0   ,0   ,0   ,0   	; $10-$1F
	.db 0x3D,0x02,0x03,0   ,0x05,0x06,0x07,0x2A,0x09,0x0A,0x1C,0   ,0   ,0   ,0   ,0x08	; $20-$2F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x36,0x35,0   ,0x0B,0x2D,0x0C	; $30-$3F

	.db 0   ,0x1F,0x32,0x30,0x21,0x13,0x22,0x23,0x24,0x18,0x25,0x26,0x27,0x34,0x33,0x19	; $40-$4F
	.db 0x1A,0x11,0x14,0x20,0x15,0x17,0x31,0x12,0x2F,0x2E,0x16,0   ,0   ,0   ,0   ,0x37	; $50-$5F
	.db 0x0D,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $60-$6F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x0F	; $70-$7F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x1b,0x28,0x29,0   ,0   ,0   ,0   ,0   	; $80-$8F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $90-$9F

IRMP_CODES_CTRL:							; ASCII Codes (Control Key)
	.db 0   ,0x1F,0x32,0x30,0x21,0x13,0x22,0x23,0x24,0x18,0x25,0x26,0x27,0x34,0x33,0x19	; $00-$0F
	.db 0x1A,0x11,0x14,0x20,0x15,0x17,0x31,0x12,0x2F,0x2E,0x16,0x09,0x0C,0x0A,0x01,0x37 ; $10-$1F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $20-$2F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $30-$3F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $40-$4F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $50-$5F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $60-$6F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0    ; $70-$7F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $80-$8F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $90-$9F

IRMP_CODES_ALT:								; ASCII Codes (Alt Key)
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $00-$0F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $10-$1F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $20-$2F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $30-$3F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $40-$4F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $50-$5F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $60-$6F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $70-$7F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $80-$8F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $90-$9F

IRMP_CODES_ALTGR:							; ASCII Codes (AltGr Key)
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $00-$0F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $10-$1F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $20-$2F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $30-$3F

	.db 0x11,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $40-$4F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x09,0x0C,0x0A,0   ,0   	; $50-$5F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $60-$6F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0x08,0x2D,0x0B,0x1C,0   	; $70-$7F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $80-$8F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   	; $90-$9F

; *****************************************************************************
; *** Main                                                                  ***
; *****************************************************************************

; (other include files are added to the end)
.include "util.inc"
.equ STR_BUF_SIZE = 80						; Size of String Buffer
.equ STR_FMT_SIZE = 16						; Size of String Format Arguments
.include "str.inc"

; ### INIT ####################################################################
init:
		cli									; Disable interrupts
		
		ldi a,LOW(RAMEND)					; Init stack pointer
		out SPL,a
		ldi a,HIGH(RAMEND)
		out SPH,a

		clr init0							; Reset init0 register
		mcall util_init						; Init Standard Modules
		mcall str_init

		mcall led_init						; Init LED Module
		mcall clk_init						; Init Clock
		mcall spi_init						; Init SPI Slave Module
		mcall irmp_init						; Init IRMP

		sei

; ### MAIN ####################################################################
main:	
		clr b
main_loop:
		mcall irmp_loop						; Scan Key Code
		
		lds a,IRMP_KEY						; Blink LED #1 on Key Press
		cpse a,zero
		ldi b,100		
		clr a
		cpse b,zero
		ser a
		mcall led_set1
		cpse b,zero
		dec b
		rjmp main_loop

; *****************************************************************************
; *** IRMP (FDC 3401 Keyboard)                                              ***
; *****************************************************************************

; ### Inits IRMP
irmp_init:
		push a
		push pdat

		in a,DDRC							; Prepare Input Pin PC5 (Data) and Output Pin PC4 (ON/OFF)
		cbr a,PINC_KEYDAT
		sbr a,PINC_ONOFF
		out DDRC,a
		in pdat,PORTC
		sbr pdat,PINC_KEYDAT
		cbr pdat,PINC_ONOFF
		out PORTC,pdat	

		inreg a,PCICR						; Config Pin PC5 for Pin Change Interrupt
		sbr a,(1<<PCIE1)
		outreg PCICR,a
		inreg a,PCMSK1
		sbr a,(1<<5)
		outreg PCMSK1,a

		sts IRMP_PULSE_TICK,zero			; Reset Pulse/Pause Flags
		sts IRMP_PULSE_COUNT,zero
		sts IRMP_PAUSE_COUNT,zero

		sts IRMP_BIT_COUNT,zero				; Reset Bit Counter for Buffer

		sts IRMP_KEY,zero					; Reset Key Code
		sts IRMP_KEY_PRESSED,zero

		sts IRMP_SHIFT,zero					; Resets Key Flags
		sts IRMP_CTRL,zero
		sts IRMP_ALT,zero
		sts IRMP_ALTGR,zero

		pop pdat
		pop a
		ret

; ### IRMP Loop
irmp_loop:
		push a
		push b
		push c
		push d
		push zl
		push zh

		lds a,CLK_TICK						; Wait for Clock Tick
		tst a
		brne irmp_loop_tick
		rjmp irmp_loop_exit

irmp_loop_tick:								; ### Clock Tick & Count Pulse/Pause
		sts CLK_TICK,zero
		lds a,IRMP_PULSE_TICK		
		tst a
		breq irmp_loop_tick_pause
irmp_loop_tick_pulse:
		sts IRMP_PULSE_TICK,zero
		lds a,IRMP_PAUSE_COUNT
		sts IRMP_PAUSE_DONE,a
		sts IRMP_PAUSE_COUNT,zero
		lds a,IRMP_PULSE_COUNT
		inc a
		cpse a,zero
		sts IRMP_PULSE_COUNT,a
		rjmp irmp_loop_analyze
irmp_loop_tick_pause:
		lds a,IRMP_PULSE_COUNT
		sts IRMP_PULSE_DONE,a
		sts IRMP_PULSE_COUNT,zero
		lds a,IRMP_PAUSE_COUNT
		inc a
		cpse a,zero
		sts IRMP_PAUSE_COUNT,a

irmp_loop_analyze:							; ### Analyze Pulse/Pause
		lds a,IRMP_PAUSE_DONE
		tst a
		brne irmp_loop_chk_stop
		rjmp irmp_loop_exit	
irmp_loop_chk_stop:							; Is Stop Bit?
		sts IRMP_PAUSE_DONE,zero
		cpi a,100							; ... Cancel on Stop Bit!
		brlo irmp_loop_chk_start
		rjmp irmp_loop_exit
irmp_loop_chk_start:	
		cpi a,16							; Is Start Bit? (> 800 us?) 
		brlo irmp_loop_chk_bit
irmp_loop_start:							; => Yes, Start Bit
		sts IRMP_BIT_COUNT,zero				; Reset Bit Counter
		rjmp irmp_loop_exit
irmp_loop_chk_bit:							; => No, Add Bit to Buffer
		clr b								; (b=Bit Value)
		cpi a,10
		brlo irmp_loop_chk_bit_cont
		ser b
irmp_loop_chk_bit_cont:						; Update Byte inside Buffer
		lds a,IRMP_BIT_COUNT				; (c=Index of Byte in Buffer)
		mov c,a
		lsr c
		lsr c
		lsr c
		ldi zl,LOW(IRMP_BUF)
		ldi zh,HIGH(IRMP_BUF)
		add zl,c
		adc zh,zero
		ld d,z								; (d=Byte to Update)
		lsr d
		cpse b,zero
		sbr d,0x80
		st z,d
		inc a								; Increment Bit Counter for Buffer
		sts IRMP_BIT_COUNT,a
		cpi a,IRMP_BITS_NUM					; All Bits Done?
		brsh irmp_loop_scan
		rjmp irmp_loop_exit

irmp_loop_scan:								; ### Scan Bits
		lds a,IRMP_BUF+IRMP_BUF_INDEX_ADDR	; Check Address Byte
		cpi a,IRMP_ADDR
		breq irmp_loop_scan_chk_press
		rjmp irmp_loop_reset
irmp_loop_scan_chk_press:					; Check Press/Release Flag
		ser d								; (d=Press Flag) 
		lds a,IRMP_BUF+IRMP_BUF_INDEX_FLAG
		cpse a,zero
		clr d
irmp_loop_scan_chk_cmd:						; Check Command
		lds a,IRMP_BUF+IRMP_BUF_INDEX_CMD	; (a=Command)
		lds b,IRMP_BUF+IRMP_BUF_INDEX_CMD_INV
		com b
		cp a,b
		breq irmp_loop_scan_chk_capslock
		rjmp irmp_loop_reset
irmp_loop_scan_chk_capslock:				; Check for "CapsLock" Key
		cpi a,IRMP_CMD_CAPSLOCK
		brne irmp_loop_scan_chk_shift
		tst d
		breq irmp_loop_scan_capslock_done
		lds a,IRMP_SHIFT
		com a
		sts IRMP_SHIFT,a
		mcall led_set2						; (Sets Led #2)					
irmp_loop_scan_capslock_done:
		rjmp irmp_loop_exit
irmp_loop_scan_chk_shift:					; Check for "Shift" Key
		cpi a,IRMP_CMD_SHIFT_LEFT
		breq irmp_loop_scan_shift
		cpi a,IRMP_CMD_SHIFT_RIGHT
		breq irmp_loop_scan_shift
		rjmp irmp_loop_scan_chk_contorl
irmp_loop_scan_shift:
		sts IRMP_SHIFT,d
		clr a								; (Resets Led #2)				
		mcall led_set2
		rjmp irmp_loop_exit
irmp_loop_scan_chk_contorl:					; Check for "Control" Key
		cpi a,IRMP_CMD_CTRL
		brne irmp_loop_scan_chk_alt
		sts IRMP_CTRL,d
		rjmp irmp_loop_exit
irmp_loop_scan_chk_alt:						; Check for "Alt" Key
		cpi a,IRMP_CMD_ALT
		brne irmp_loop_scan_chk_altgr
		sts IRMP_ALT,d
		rjmp irmp_loop_exit
irmp_loop_scan_chk_altgr:					; Check for "AltGr" Key
		cpi a,IRMP_CMD_ALTGR
		brne irmp_loop_scan_key
		sts IRMP_ALTGR,d
		rjmp irmp_loop_exit

irmp_loop_scan_key:							; ### Scan Key		
		tst d								; Is Key Pressed?
		brne irmp_loop_scan_key0
		sts IRMP_KEY,zero					; => No, Reset Key Code and Cancel...
		rjmp irmp_loop_exit
irmp_loop_scan_key0:						; => Yes, Set Pointer for Table Key Codes (ASCII Code)
		ldi zl,LOW(2*IRMP_CODES)			
		ldi zh,HIGH(2*IRMP_CODES)
		lds b,IRMP_SHIFT
		tst b
		breq irmp_loop_scan_key1
		ldi zl,LOW(2*IRMP_CODES_SHIFT)
		ldi zh,HIGH(2*IRMP_CODES_SHIFT)
		rjmp irmp_loop_scan_key_search
irmp_loop_scan_key1:
		lds b,IRMP_CTRL
		tst b
		breq irmp_loop_scan_key2
		ldi zl,LOW(2*IRMP_CODES_CTRL)
		ldi zh,HIGH(2*IRMP_CODES_CTRL)
		rjmp irmp_loop_scan_key_search
irmp_loop_scan_key2:
		lds b,IRMP_ALT
		tst b
		breq irmp_loop_scan_key3
		ldi zl,LOW(2*IRMP_CODES_ALT)
		ldi zh,HIGH(2*IRMP_CODES_ALT)
		rjmp irmp_loop_scan_key_search
irmp_loop_scan_key3:
		lds b,IRMP_ALTGR
		tst b
		breq irmp_loop_scan_key_search
		ldi zl,LOW(2*IRMP_CODES_ALTGR)
		ldi zh,HIGH(2*IRMP_CODES_ALTGR)
		rjmp irmp_loop_scan_key_search
irmp_loop_scan_key_search:		
		clr c								; (c=Counter for Key Code, Index inside Table)
irmp_loop_scan_key_search_loop:
		lpm b,z+
		cp a,b
		breq irmp_loop_key_code
		inc c
		cpi c,160
		brne irmp_loop_scan_key_search_loop
		sts IRMP_KEY,zero					; Code Not found, Reset and Cancel...
		rjmp irmp_loop_exit
irmp_loop_key_code:							; ### Valid Key Code
		sts IRMP_KEY,c						; Store Key Code (ASCII Code)
		sts IRMP_KEY_PRESSED,c				
		cpi c,IRMP_KEY_CODE_ONOFF			; Is Special ON/OFF Key Pressed?
		brne irmp_loop_key_code_done
		mcall irmp_pulse_onoff				; => Yes, Pulse ON/OFF Pin for 100ms
irmp_loop_key_code_done:		
		rjmp irmp_loop_exit

irmp_loop_reset:							; Resets Key 
		sts IRMP_KEY,zero					
		sts IRMP_SHIFT,zero
		sts IRMP_CTRL,zero
		sts IRMP_ALT,zero
		sts IRMP_ALTGR,zero

irmp_loop_exit:
		pop zh
		pop zl
		pop d
		pop c
		pop b
		pop a
		ret

; ### Pulse Pin for ON/OFF (100ms)
irmp_pulse_onoff:
		push a
		push pdat

		in pdat,PORTC
		sbr pdat,PINC_ONOFF
		out PORTC,pdat	
		ldi a,100
		mcall util_delay_Xms
		cbr pdat,PINC_ONOFF
		out PORTC,pdat	

		pop pdat
		pop a
		ret

; ### IRMP "Pin Change #1" (PC5/PCINT13)
irmp_isr_pulse:
		push a
		in stat,SREG

		ser a								; Set Pulse Flag
		sts IRMP_PULSE_TICK,a

irmp_isr_pulse_exit:
		out SREG,stat
		pop a
		reti


; *****************************************************************************
; *** Clock                                                                ***
; *****************************************************************************

; ### Inits Timer
clk_init:
		push a

		clr a								; Init Timer1
		sts TCCR1A,a
		ldi a,(1<<WGM12)|CLK_PRESCALER	
		sts TCCR1B,a						
		ldi a,HIGH(CLK_PRELOAD)				
		sts OCR1AH,a
		ldi a,LOW(CLK_PRELOAD)
		sts OCR1AL,a
		lds a,TIMSK1
		sbr a,(1<<OCIE1A)
		sts TIMSK1,a

		sts CLK_TICK,zero					; Reset Tick Flag

		pop a
		ret

; ### Interrupt "Output Compare1A" 
clk_isr_timer1:
		push a
		in stat,SREG
		
		ser a								; Set Tick Flag
		sts CLK_TICK,a

		out SREG,stat
		pop a
		reti

; *****************************************************************************
; *** LED                                                                   ***
; *****************************************************************************

; ### Inits LED
led_init:
		push a

		in a,DDRB							; Set LED Pins to Output
		sbr a,PINB_LED1|PINB_LED2
		out DDRB,a
		
		clr a								; Disable LEDs
		mcall led_set1						
		mcall led_set2

		pop a
		ret

; ### Switch LED #1
; >a: The Flag
led_set1:
		push pdat

		in pdat,PORTB
		cbr pdat,PINB_LED1
		cpse a,zero
		sbr pdat,PINB_LED1
		out PORTB,pdat

		pop pdat
		ret

; ### Switch LED #2
; >a: The Flag
led_set2:
		push pdat

		in pdat,PORTB
		cbr pdat,PINB_LED2
		cpse a,zero
		sbr pdat,PINB_LED2
		out PORTB,pdat

		pop pdat
		ret

; *****************************************************************************
; *** SPI (Serial Slave)                                                    ***
; *****************************************************************************
spi_init:
		push a
		push pdat

		in a,DDRB							; Prepare pins
		cbr a,PINB_CSN|PINB_MOSI|PINB_SCK
		sbr a,PINB_MISO
		out DDRB,a
		in pdat,PORTB
		sbr pdat,PINB_CSN|PINB_MOSI|PINB_SCK
		cbr pdat,PINB_MISO
		out PORTB,pdat	
		
		clr a								; Configure SPI
		sbr a,(1<<SPIE)						; SPIE (SPI interrupt enable)
		sbr a,(1<<SPE)						; SPE (SPI enable)
		sbr a,(0<<DORD)						; DORD (Data order)				-> 0: MSB of the data word is transmitted first / 1: LSB of the data word is transmitted first.
		sbr a,(0<<MSTR)						; MSTR (Master/slave select)	-> 0: Slave / 1: Master 
		sbr a,(0<<CPOL)						; CPOL (Clock polarity)			-> 0: Leading edge = Rising + Trailing edge = Falling (SCK low when idle) / 1: Leading edge = Falling + Trailing edge = Rising
		sbr a,(0<<CPHA)						; CPHA (Clock phase)			-> 0: Leading edge = Sample + Trailing edge = Setup / 1: Leading edge = Setup + Trailing edge = Sample
		sbr a,(0<<SPR0)|(0<<SPR1)			; SPRx (SPI clock rate select)	-> 000: fosc/4, 001: fosc/16, 010: fosc/64, 011: fosc/128, 100: fosc/2, 101: fosc/8, 110: fosc/32, 111: fosc/64
		out SPCR,a

		clr a
		cbr a,(0<<SPI2X)					; SPI2X: Disable Double SPI speed bit (see SPRx)
		out SPSR,a

		pop pdat
		pop a
		ret

; ### ISR "SPI Complete"
spi_isr_data:
		push a
		in stat,SREG

		lds a,IRMP_KEY_PRESSED
		out SPDR,a
		sts IRMP_KEY_PRESSED,zero

spi_isr_data_exit:
		out SREG,stat
		pop a
		reti
