; LCDControllerVT52
; (C) Mario Latzig 2015
;
; on ATmega328P
;
; PB0: LED    (External LED)
; PB1: BUSY   (Busy Flag)
; PB2: CSN    (HW SPI)
; PB3: MOSI	  (HW SPI)
; PB4: MISO   (HW SPI) 
; PB5: SCK    (HW SPI)
;
; PC0: IO0	  (External I/O)
; PC1: IO1	  (External I/O)
; PC2: IO2	  (External I/O)
; PC3: IO3	  (External I/O)
; PC4: IO4	  (External I/O)
; PC5: IO5	  (External I/O)
;
; PD0: DIO0   (LCD)
; PD1: DIO1   (LCD)
; PD2: DIO2   (LCD)
; PD3: DIO3   (LCD)
; PD4: CL1    (LCD)
; PD5: CL2    (LCD)
; PD6: FLM    (LCD)
; PD7: DISP   (LCD)

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************
.equ FC = 24000000							; Clock = 24 MHz (overclocked!)

.def temp = r0								; Temporary Register
.def zero = r1								; Always Zero
.def statreg = r2							; Backup of Status Register inside ISR
.def blink = r3								; Blinking Cursor
.def curen = r4								; Enabled Flag for Cursor
.def curx = r6								; Cursor X-Position
.def cury = r7								; Cursor Y-Position
.def curbx = r8								; Cursor X-Position (Backup)
.def curby = r9								; Cursor Y-Position (Backup)
.def fg = r10								; Foreground
.def bg = r11								; Background
.def nl = r12								; Flag for New Line
.def gr = r13								; Flag for Graphics Mode
.def row = r16								; Current Row Counter for LCD
.def a = r17								; Temporary Register (a) 
.def b = r18								; Temporary Register (b)
.def c = r19								; Temporary Register (c) 
.def d = r20								; Temporary Register (d) 
.def e = r21								; Temporary Register (e)
.def f = r22								; Temporary Register (f)
.def pdat = r23								; Remporary Port Register
.def ol = r24								; Counter Register (LO)
.def oh = r25								; Counter Register (HI)

; ### PINS ####################################################################

.equ PINB_LED = (1<<0)						; PB0: LED    (External LED)
.equ PINB_BUSY = (1<<1)						; PB1: BUSY   (Busy Flag)
.equ PINB_CSN = (1<<2)						; PB2: CSN    (HW SPI)
.equ PINB_MOSI = (1<<3)						; PB3: MOSI	  (HW SPI)
.equ PINB_MISO = (1<<4)						; PB4: MISO   (HW SPI) 
.equ PINB_SCK = (1<<5)						; PB5: SCK    (HW SPI)

.equ PIND_DIO0 = (1<<0)						; PD0: DIO0   (LCD Data Nibble Bit #0)
.equ PIND_DIO1 = (1<<1)						; PD1: DIO1   (LCD Data Nibble Bit #1)
.equ PIND_DIO2 = (1<<2)						; PD2: DIO2   (LCD Data Nibble Bit #2)
.equ PIND_DIO3 = (1<<3)						; PD3: DIO3   (LCD Data Nibble Bit #3)
.equ PIND_CL1 = (1<<4)						; PD4: CL1    (LCD Line HSync)
.equ PIND_CL2 = (1<<5)						; PD5: CL2    (LCD Pixel Clock)
.equ PIND_FLM = (1<<6)						; PD6: FLM    (LCD Frame VSync)
.equ PIND_DISP = (1<<7)						; PD7: DISP   (LCD Display Enable)

; === LCD (Display) ===========================================================
.equ LCD_WIDTH = 640						; Number of Pixels Horizontal 
.equ LCD_HEIGHT = 200						; Number of Pixels Vertical
.equ LCD_NIBBLES = LCD_WIDTH / 4			; Number of Nibbles Horizontal
.equ LCD_COLS = LCD_WIDTH / 8				; Number of Columns per Row (80 Chars)
.equ LCD_ROWS = LCD_HEIGHT / 8				; Number of Rows (25 Rows)
.equ LCD_LINES = LCD_ROWS - 1				; Number of Rows for Console (24 Rows)
.equ LCD_TXT_SIZE = LCD_COLS * LCD_LINES	; Size of Text Buffer
.equ LCD_BUF_SIZE = LCD_COLS * LCD_ROWS		; Size of Text Buffer (including Status Line)

; Status Line
.equ LCD_STAT_ROW = LCD_LINES				; Row for Status Line inside Text Buffer
.equ LCD_STAT_OFF = LCD_STAT_ROW*LCD_COLS	; Offset inside Text Buffer for Status Line
.equ LCD_STAT_COL_GR = 39					; Status Line Column: Graphics Character Flag
.equ LCD_STAT_COL_NL = 48					; Status Line Column: New Line Enabled Flag
.equ LCD_STAT_COL_REVERSE = 58				; Status Line Column: Reverse Enabled Flag
.equ LCD_STAT_COL_ENABLED = 68				; Status Line Column: Cursor Enabled Flag
.equ LCD_STAT_COL_XPOS = 73					; Status Line Column: Cursor X-Position
.equ LCD_STAT_COL_YPOS = 77					; Status Line Column: Cursor Y-Position

; Escape Buffer
.equ LCD_ESC_BUF_SIZE = 4					; Size of Buffer for Escape Sequences

; *****************************************************************************
; *** Macros                                                                ***
; *****************************************************************************

; ### Adds Byte (1) to Register (0)
.macro adi 
	subi @0, -(@1)
.endm

; ### Adds Word (1) to Register Pair (0H:0L)
.macro addiw
    subi @0L, LOW(-@1)  
    sbci @0H, HIGH(-@1)
.endm

; ### Perform call (0) to other module
.macro mcall 
.if (FLASHEND > 0x0fff)
   call @0
.else
   rcall @0
.endif
.endm

.eseg
; *****************************************************************************
; *** EEPROM                                                                ***
; *****************************************************************************

	.db 0	

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

; === LCD (Display) ===========================================================
LCD_BUF:		.byte LCD_BUF_SIZE			; Text Buffer (80 * 24 Chars)
LCD_ESC_LEN:	.byte 1						; Number of Chars for Escape Sequence
LCD_ESC_BUF:	.byte LCD_ESC_BUF_SIZE		; Buffer for Escape Sequences (4 Chars)

.cseg
; *****************************************************************************
; *** Vectors                                                               ***
; *****************************************************************************

.org 0x0000
		jmp init
.org SPIaddr								; Interrupt for Serial Transfer (Receive)
		jmp spi_recv_intr									
.org INT_VECTORS_SIZE

; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

.include "font.inc"							; FONT array in program memory

; === LCD (Display) ===========================================================
LCD_BUF_OFFSETS:							; Offset of Rows inside Text Buffer
	.dw 0*LCD_COLS,1*LCD_COLS,2*LCD_COLS,3*LCD_COLS,4*LCD_COLS,5*LCD_COLS,6*LCD_COLS,7*LCD_COLS
	.dw 8*LCD_COLS,9*LCD_COLS,10*LCD_COLS,11*LCD_COLS,12*LCD_COLS,13*LCD_COLS,14*LCD_COLS,15*LCD_COLS
	.dw 16*LCD_COLS,17*LCD_COLS,18*LCD_COLS,19*LCD_COLS,20*LCD_COLS,21*LCD_COLS,22*LCD_COLS,23*LCD_COLS
	.dw 24*LCD_COLS
LCD_STATUS_TXT:								; Format of Status Row (80 Chars)
	.db "                                  |Gr: Off |NL: On  |Rev: Off |Cur: On  X00 Y00 "
LCD_ESC_COMMANDS:							; Enumeration of ESC Commands, see Jump Table "lcd_esc_table"
	.db 'A','B','C','D','E','F','G','H','I','J','K','L','M','Y','b','c','d','e','f','j','k','l','o','p','q','v','w',0

; *****************************************************************************
; *** Main                                                                  ***
; *****************************************************************************

; ### INIT ####################################################################
init:
		cli									; Disable interrupts
		
		clr zero							
		ldi a,LOW(RAMEND)					; Init stack pointer
		out SPL,a
		ldi a,HIGH(RAMEND)
		out SPH,a

		mcall spi_init						; Inits HW SPI
		mcall lcd_init						; Inits LCD

		sei									

; #############################################################################
; ### FRAME LOOP ##############################################################
; #############################################################################

frame:
		clr row								; Reset Row Counter

		in pdat,PORTD						; Read Port	

		sbr pdat,PIND_DISP					; Enable Display
		out PORTD,pdat

		ldi xl,LOW(LCD_BUF)					; Set X-Pointer to Start of Text Buffer
		ldi xh,HIGH(LCD_BUF)		

; #############################################################################
; ### LINE LOOP ###############################################################
; #############################################################################

line:
		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Row for Text Buffer 
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,row
		lsr a
		lsr a
		lsr a
		mov c,a								; (c = Current Text Row: 0-24)
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi xl,LOW(LCD_BUF)
		ldi xh,HIGH(LCD_BUF)
		add xl,ol
		adc xh,oh
		mov d,row							; (d = Current Font Row: 0-7)
		andi d,0x07
		clr b								; (b = Current Text Column: 0-79)

		clr f								; Should show Cursor (Right Row)?
		tst curen							; (f = Cursor Blink Flag for Row)
		breq line_loop
		cpse c,cury							
		rjmp line_loop
		sbrc blink,5
		rjmp line_loop
		ser f								

line_loop:
		ld a,x+								; Get Console Char & Get Font Byte 
		ldi zl,LOW(2*FONT)					
		ldi zh,HIGH(2*FONT)
		add zl,a
		adc zh,d
		lpm a,z

		cpse b,curx							; Should show Cursor (Right Column)?
		rjmp line_loop_draw
		eor a,f

line_loop_draw:
		; MSB Nibble of Char
		cbr pdat,PIND_DIO0|PIND_DIO1|PIND_DIO2|PIND_DIO3
		mov e,a
		swap e 
		andi e,0x0F
		or pdat,e
		out PORTD,pdat
		sbr pdat,PIND_CL2					; CL2 = 1 (Pixel Clock)
		out PORTD,pdat
		cbr pdat,PIND_CL2					; CL2 = 0
		out PORTD,pdat

		; LSB Nibble of Char
		cbr pdat,PIND_DIO0|PIND_DIO1|PIND_DIO2|PIND_DIO3
		andi a,0x0F
		or pdat,a
		out PORTD,pdat
		sbr pdat,PIND_CL2					; CL2 = 1 (Pixel Clock)
		out PORTD,pdat
		cbr pdat,PIND_CL2					; CL2 = 0
		out PORTD,pdat

		inc b								; Increment Column
		cpi b,LCD_COLS						; End of Line?
		brne line_loop
		
		tst row
		brne line_next
		sbr pdat,PIND_FLM					; FLM = 1 (New Frame)
		out PORTD,pdat
line_next:				
		sbr pdat,PIND_CL1					; Clock CL1 (Row Clock)
		out PORTD,pdat
		cbr pdat,PIND_CL1|PIND_FLM			; FLM = 0
		out PORTD,pdat

; #############################################################################
; ### LINE POST ###############################################################
; #############################################################################

line_post:

; #############################################################################
; ### FRAME POST ##############################################################
; #############################################################################

frame_post:
		inc row								; Increment Row
		cpi row,LCD_HEIGHT
		breq frame_next
		rjmp line

frame_next:
		cbr pdat,PIND_DISP					; Disable Display
		out PORTD,pdat

		inc blink							; Increment Blinker for Cursor

frame_status_line:							; ### Updates Status Line
frame_status_line_gr_en:					; Graphics Character Flag
		tst gr
		breq frame_status_line_gr_dis
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_GR+0,a
		ldi a,'n'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_GR+1,a
		ldi a,' '+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_GR+2,a
		rjmp frame_status_line_rev_en
frame_status_line_gr_dis:
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_GR+0,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_GR+1,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_GR+2,a
frame_status_line_nl_en:					; New Line Flag
		tst nl
		breq frame_status_line_nl_dis
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_NL+0,a
		ldi a,'n'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_NL+1,a
		ldi a,' '+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_NL+2,a
		rjmp frame_status_line_rev_en
frame_status_line_nl_dis:
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_NL+0,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_NL+1,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_NL+2,a
frame_status_line_rev_en:					; Reverse Enabled Flag
		tst bg
		breq frame_status_line_rev_dis
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_REVERSE+0,a
		ldi a,'n'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_REVERSE+1,a
		ldi a,' '+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_REVERSE+2,a
		rjmp frame_status_line_cur_en
frame_status_line_rev_dis:
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_REVERSE+0,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_REVERSE+1,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_REVERSE+2,a
frame_status_line_cur_en:					; Cursor Enabled Flag
		tst curen
		breq frame_status_line_cur_dis
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_ENABLED+0,a
		ldi a,'n'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_ENABLED+1,a
		ldi a,' '+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_ENABLED+2,a
		rjmp frame_status_line_cur_pos
frame_status_line_cur_dis:
		ldi a,'O'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_ENABLED+0,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_ENABLED+1,a
		ldi a,'f'+0x60
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_ENABLED+2,a
frame_status_line_cur_pos:
		mov a,curx							; Cursor X-Position
		inc a
		mcall util_byte2dec_s
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_XPOS+0,b
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_XPOS+1,c
		mov a,cury							; Cursor Y-Position
		inc a
		mcall util_byte2dec_s
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_YPOS+0,b
		sts LCD_BUF+LCD_STAT_OFF+LCD_STAT_COL_YPOS+1,c
		
		;rcall cmd_perform					; Get & Perform Command State
		rjmp frame

; *****************************************************************************
; *** Display                                                               ***
; *****************************************************************************

; ### Inits Display
lcd_init:
		push a
		push b
		push pdat
		push xl
		push xh
		push zl
		push zh

		in a,DDRD							; Set LCD Pins as Output
		sbr a,PIND_DIO0|PIND_DIO1|PIND_DIO2|PIND_DIO3|PIND_CL1|PIND_CL2|PIND_FLM|PIND_DISP 
		out DDRD,a
		in pdat,PORTD
		cbr pdat,PIND_DIO0|PIND_DIO1|PIND_DIO2|PIND_DIO3|PIND_CL1|PIND_CL2|PIND_FLM|PIND_DISP
		out PORTD,pdat

		mcall lcd_clear						; Clears Display
		mcall lcd_cursor_enable				; Enable Cursor
		mcall lcd_esc_reset					; Resets Buffer for Escape Sequence

		clr curbx							; Resets Backup of Cursor Position
		clr curby

		ser a								; Presets Color
		mov fg,a
		clr bg

		ser a								; Enable New Line (at EOL)
		mov nl,a

		clr gr								; Disable Graphic Characters Mode

lcd_init_status:							; ### Inits Status Line
		ldi zl,LOW(2*LCD_STATUS_TXT)
		ldi zh,HIGH(2*LCD_STATUS_TXT)
		ldi xl,LOW(LCD_BUF+LCD_STAT_OFF)
		ldi xh,HIGH(LCD_BUF+LCD_STAT_OFF)
		ldi b,LCD_COLS
lcd_init_status_loop:
		lpm a,z+
		sbr a,0x80
		subi a,' '
		st x+,a
		dec b 
		brne lcd_init_status_loop

		pop zh
		pop zl
		pop xh
		pop xl
		pop pdat
		pop b
		pop a
		ret

; ### Clears Display
lcd_clear:
		push a
		push ol
		push oh
		push zl
		push zh

		ldi ol,LOW(LCD_TXT_SIZE)
		ldi oh,HIGH(LCD_TXT_SIZE)
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
lcd_clear_loop:
		st z+,zero
		sbiw ol:oh,1
		brne lcd_clear_loop

		clr blink
		mcall lcd_home

		pop zh
		pop zl
		pop oh
		pop ol
		pop a
		ret

; ### Clear Screen from Start
lcd_clear_start:
		push a
		push ol
		push oh
		push zl 
		push zh

		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Number of Chars
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		add ol,curx
		adc oh,zero
		adiw ol:oh,1
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)

lcd_clear_start_loop:						; Perform Delete
		st z+,zero
		sbiw ol:oh,1
		brne lcd_clear_start_loop

		pop zh
		pop zl
		pop oh
		pop ol
		pop a
		ret

; ### Clear Remaining Screen
lcd_clear_end:
		push a
		push ol
		push oh
		push zl 
		push zh

		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer (Current Position)
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		add zl,curx
		adc zh,zero
		add ol,curx							; Calculates Number of Chars
		adc oh,zero
		com oh
		neg ol
		sbci oh,-1
		addiw o,LCD_TXT_SIZE

lcd_clear_end_loop:							; Perform Delete
		st z+,zero
		sbiw ol:oh,1
		brne lcd_clear_end_loop

		pop zh
		pop zl
		pop oh
		pop ol
		pop a
		ret

; ### Delete Complete Line
lcd_clear_line:
		push a
		push ol
		push oh
		push zl 
		push zh

		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer (Current Position)
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		ldi a,LCD_COLS
		mcall lcd_start						; Set Cursor at the Beginning of the Line

lcd_clear_line_loop:						; Perform Delete
		st z+,zero
		dec a
		brne lcd_clear_line_end_loop

		pop zh
		pop zl
		pop oh
		pop ol
		pop a
		ret

; ### Delete Beginning of Line
lcd_clear_line_start:
		push a
		push ol
		push oh
		push zl 
		push zh

		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer (Current Position)
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		mov a,curx							; Calculates Number of Chars
		inc a

lcd_clear_line_start_loop:					; Perform Delete
		st z+,zero
		dec a
		brne lcd_clear_line_end_loop

		pop zh
		pop zl
		pop oh
		pop ol
		pop a
		ret

; ### Delete Rest of Line
lcd_clear_line_end:
		push a
		push ol
		push oh
		push zl 
		push zh

		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer (Current Position)
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		add zl,curx
		adc zh,zero
		mov a,curx							; Calculates Number of Chars
		neg a
		adi a,LCD_COLS

lcd_clear_line_end_loop:					; Perform Delete
		st z+,zero
		dec a
		brne lcd_clear_line_end_loop

		pop zh
		pop zl
		pop oh
		pop ol
		pop a
		ret

; ### Home Position
lcd_home:
		push b
		push c
		
		clr b
		clr c
		mcall lcd_cursor_set

		pop c
		pop b
		ret

; ### Get Char from Buffer
; <a: The Char
; >b: The X-Position
; >c: The Y-Position
lcd_get:
		push ol
		push oh
		push zl
		push zh

		clr a								; Check Range of Positions
		cpi b,LCD_COLS
		brsh lcd_get_exit
		cpi c,LCD_ROWS
		brsh lcd_get_exit

		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,c
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		add zl,b
		adc zh,zero
		ld a,z								; Gets Char from Buffer

lcd_get_exit:	
		pop zh
		pop zl
		pop oh
		pop ol
		ret

; ### Set Char in Buffer
; >a: The Char
; >b: The X-Position
; >c: The Y-Position
lcd_set:
		push ol
		push oh
		push zl
		push zh

		cpi b,LCD_COLS						; Check Range of Positions
		brsh lcd_set_exit
		cpi c,LCD_ROWS
		brsh lcd_set_exit

		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer 
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov ol,c
		lsl ol
		add zl,ol
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		add zl,b
		adc zh,zero
		st z,a

lcd_set_exit:	
		pop zh
		pop zl
		pop oh
		pop ol
		ret

; ### Enablde Cursor
lcd_cursor_enable:
		push a

		ser a
		mov curen,a

		pop a
		ret

; ### Disable Cursor
lcd_cursor_disable:
		clr curen
		ret

; ### Set Cursor
; >b: The X-Position
; >c: The Y-Position
lcd_cursor_set:
		push b
		push c

		cpi b,LCD_COLS						; Check Range of Positions
		brlo lcd_cursor_set_1
		ldi b,LCD_COLS-1
lcd_cursor_set_1:
		cpi c,LCD_LINES
		brlo lcd_cursor_set_2
		ldi c,LCD_LINES-1
lcd_cursor_set_2:

		mov curx,b
		mov cury,c
		clr blink

		pop c
		pop b
		ret

; ### Add Char
; >a: The ASCII Character
lcd_add:
		push a
		push b
		push c
		
		cpi a,' '							; Is ASCII Char?
		brlo lcd_add_space
		cpi a,0x80
		brsh lcd_add_space
		subi a,' '							
		tst gr								; Is Graphic enabled?
		breq lcd_add_chk_color
		cpi a,0x40
		brlo lcd_add_chk_color
		adi a,0x20
		rjmp lcd_add_chk_color
lcd_add_space:		
		clr a
lcd_add_chk_color:							; Check Color
		cp fg,bg						
		brne lcd_add_chk_reverse
		clr a
lcd_add_chk_reverse:						; Check Inverse Text
		cpse bg,zero
		sbr a,0x80		
lcd_add_store:								; Store Char in Text Buffer
		mov b,curx
		mov c,cury
		mcall lcd_set
		mcall lcd_next		

		pop c
		pop b
		pop a
		ret

; ### Delete Last Char
lcd_delete:
		push a
		push b
		push c
		
		tst curx							; Decrement Cursor X-Position (if possible)
		breq lcd_delete_exit
		dec curx
		clr a								; Delete Char 
		mov b,curx
		mov c,cury
		mcall lcd_set

lcd_delete_exit:
		pop c
		pop b
		pop a
		ret

; ### Cursor: Next Position
lcd_next:
		push a
		
		clr blink		
		inc curx							; Increment Cursor X-Position
		mov a,curx
		cpi a,LCD_COLS						; End Of Line?
		brlo lcd_next_exit
		tst nl								; Should do New Line?
		brne lcd_next_new_line
		ldi a,LCD_COLS-1					; => No, stays at EOL
		mov curx,a
		rjmp lcd_next_exit
lcd_next_new_line:							; => Yes, do New Line
		mcall lcd_new_line

lcd_next_exit:
		pop a
		ret


; ### Cursor: Start Line
lcd_start:
		clr blink
		clr curx
		ret

; ### Cursor: New Line
lcd_new_line:
		push a

		clr blink
		clr curx
		inc cury
		mov a,cury
		cpi a,LCD_LINES
		brlo lcd_new_line_exit
		ldi a,LCD_LINES-1
		mov cury,a
		mcall lcd_scroll_up

lcd_new_line_exit:
		pop a
		ret

; ### Scroll Up
lcd_scroll_up:
		push a
		push ol
		push oh
		push xl
		push xh
		push zl
		push zh

		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		ldi xl,LOW(LCD_BUF+LCD_COLS)
		ldi xh,HIGH(LCD_BUF+LCD_COLS)
		ldi ol,LOW(LCD_TXT_SIZE-LCD_COLS)
		ldi oh,HIGH(LCD_TXT_SIZE-LCD_COLS)
lcd_scroll_up_loop:
		ld a,x+
		st z+,a
		sbiw ol:oh,1
		brne lcd_scroll_up_loop
lcd_scroll_up_clear:
		ldi a,LCD_COLS
lcd_scroll_up_clear_loop:
		st z+,zero
		dec a
		brne lcd_scroll_up_clear_loop

		pop zh
		pop zl
		pop xh
		pop xl
		pop oh
		pop ol
		pop a
		ret

; ### Move Up (Cursor Up)
lcd_move_up:
		push a
		push ol
		push oh
		push xl
		push xh
		push zl
		push zh

		tst cury							; Is already first Line?
		breq lcd_move_up_scroll
		cpse cury,zero						; Move Cursor Up
		dec cury
		clr blink
		rjmp lcd_move_up_exit
	
lcd_move_up_scroll:							; ### SCROLL DOWN
		ldi zl,LOW(LCD_BUF+LCD_TXT_SIZE)	
		ldi zh,HIGH(LCD_BUF+LCD_TXT_SIZE)
		ldi xl,LOW(LCD_BUF+LCD_TXT_SIZE-LCD_COLS)
		ldi xh,HIGH(LCD_BUF+LCD_TXT_SIZE-LCD_COLS)
		ldi ol,LOW(LCD_TXT_SIZE-LCD_COLS)
		ldi oh,HIGH(LCD_TXT_SIZE-LCD_COLS)
lcd_move_up_scroll_loop:					; Perform Scroll
		ld a,-x
		st -z,a
		sbiw ol:oh,1
		brne lcd_move_up_scroll_loop
lcd_move_up_clear:							; ### DELETE TOPMOST LINE 
		ldi zl,LOW(LCD_BUF)	
		ldi zh,HIGH(LCD_BUF)
		ldi a,LCD_COLS
lcd_move_up_clear_loop:
		st z+,zero
		dec a
		brne lcd_move_up_clear_loop

lcd_move_up_exit:
		pop zh
		pop zl
		pop xh
		pop xl
		pop oh
		pop ol
		pop a
		ret		

; ### Inserts Line
lcd_line_insert:
		push a
		push ol
		push oh
		push xl
		push xh
		push zl
		push zh

		mov a,cury							; Is already Last Line?
		cpi a,LCD_LINES-1
		brsh lcd_line_insert_clear
lcd_line_insert_scroll:						; ### SCROLL DOWN
		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Number of Chars
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		com oh								
		neg ol
		sbci oh,-1
		addiw o,LCD_TXT_SIZE
		ldi zl,LOW(LCD_BUF+LCD_TXT_SIZE)	; Set Addr of Last Line
		ldi zh,HIGH(LCD_BUF+LCD_TXT_SIZE)
		ldi xl,LOW(LCD_BUF+LCD_TXT_SIZE-LCD_COLS)
		ldi xh,HIGH(LCD_BUF+LCD_TXT_SIZE-LCD_COLS)
lcd_line_insert_scroll_loop:				; Perform Scroll
		ld a,-x
		st -z,a
		sbiw ol:oh,1
		brne lcd_line_insert_scroll_loop

lcd_line_insert_clear:						; ### DELETE CURRENT LINE
		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer (Current Position)
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		ldi a,LCD_COLS
lcd_line_insert_clear_loop:
		st z+,zero
		dec a
		brne lcd_line_insert_clear_loop

		mcall lcd_start						; Set Cursor at the Beginning of the Line

		pop zh
		pop zl
		pop xh
		pop xl
		pop oh
		pop ol
		pop a
		ret

; ## Deletes Line
lcd_line_delete:
		push a
		push ol
		push oh
		push xl
		push xh
		push zl
		push zh

		mov a,cury							; Is already Last Line?
		cpi a,LCD_LINES-1
		brsh lcd_line_delete_clear
lcd_line_delete_scroll:						; ### SCROLL
		ldi zl,LOW(2*LCD_BUF_OFFSETS)		; Calculates Addr for Text Buffer (Current Line)
		ldi zh,HIGH(2*LCD_BUF_OFFSETS)
		mov a,cury
		lsl a
		add zl,a
		adc zh,zero
		lpm ol,z+
		lpm oh,z
		ldi zl,LOW(LCD_BUF)
		ldi zh,HIGH(LCD_BUF)
		add zl,ol
		adc zh,oh
		mov xl,zl							; Calulcates Addr for Next Line
		mov xh,zh
		addiw x,LCD_COLS
		com oh								; Calculates Number of Chars
		neg ol
		sbci oh,-1
		addiw o,LCD_TXT_SIZE
lcd_line_delete_scroll_loop:				; Perform Scroll
		ld a,x+
		st z+,a
		sbiw ol:oh,1
		brne lcd_line_delete_scroll_loop

lcd_line_delete_clear:						; ### DELETE LAST LINE
		ldi zl,LOW(LCD_BUF+LCD_TXT_SIZE-LCD_COLS)
		ldi zh,HIGH(LCD_BUF+LCD_TXT_SIZE-LCD_COLS)
		add zl,ol
		adc zh,oh
		ldi a,LCD_COLS
lcd_line_delete_clear_loop:
		st z+,zero
		dec a
		brne lcd_line_delete_clear_loop

		pop zh
		pop zl
		pop xh
		pop xl
		pop oh
		pop ol
		pop a
		ret

; ### Reset Buffer for Escape Sequence
lcd_esc_reset:
		push a
		push zl
		push zh

		clr a								; Reset Length of Buffer
		sts LCD_ESC_LEN,a

		ldi zl,LOW(LCD_ESC_BUF)				; Clears Buffer
		ldi zh,HIGH(LCD_ESC_BUF)
		ldi a,LCD_ESC_BUF_SIZE
lcd_esc_reset_loop:
		st z+,zero
		dec a
		brne lcd_esc_reset_loop

		pop zh
		pop zl
		pop a
		ret

; ### Adds Char to Escape Sequence
; >a: The Char
lcd_esc_add:
		push temp
		push a
		push zl
		push zh
		
		mov temp,a							; Out of Range for Buffer?
		lds a,LCD_ESC_LEN
		cpi a,LCD_ESC_BUF_SIZE
		brlo lcd_esc_add_save
		mcall lcd_esc_reset					; => Yes, Reset Buffer...
		rjmp lcd_esc_add_exit
lcd_esc_add_save:							; => No, Add Char to Buffer
		ldi zl,LOW(LCD_ESC_BUF)
		ldi zh,HIGH(LCD_ESC_BUF)
		add zl,a
		adc zh,zero
		st z+,temp
		inc a
		sts LCD_ESC_LEN,a

lcd_esc_add_exit:
		pop zh
		pop zl
		pop a
		pop temp
		ret

; ### Gets Length of Escape Sequence
; <a: The Length
lcd_esc_get_len:
		lds a,LCD_ESC_LEN
		ret

; ### Performs Escape Sequence inside Buffer
lcd_esc_perform:
		push a
		push b
		push c
		push zl
		push zh

		lds a,LCD_ESC_LEN					; Length >= 2?
		cpi a,2
		brlo lcd_esc_perform_exit
		
		lds c,LCD_ESC_BUF+1					; Get Command Char
		ldi zl,LOW(2*LCD_ESC_COMMANDS)
		ldi zh,HIGH(2*LCD_ESC_COMMANDS)
		clr b
lcd_esc_perform_find:						; Find Index for Jump Table
		lpm a,z+
		tst a
		breq lcd_esc_perform_reset
		cp a,c
		breq lcd_esc_perform_jump
		inc b
		rjmp lcd_esc_perform_find

lcd_esc_perform_jump:						; Perform Jump
		ldi zl,LOW(lcd_esc_table)
		ldi zh,HIGH(lcd_esc_table)
		add zl,b
		adc zh,zero
		icall
		rjmp lcd_esc_perform_exit

lcd_esc_perform_reset:						
		mcall lcd_esc_reset

lcd_esc_perform_exit:
		pop zh
		pop zl
		pop c
		pop b
		pop a
		ret

; ==== ESCAPE COMMANDS ========================================================
lcd_esc_table:
		rjmp lcd_esc_perform_up				; ESC A (Cursor Up)
		rjmp lcd_esc_perform_down			; ESC B (Cursor Down)
		rjmp lcd_esc_perform_right			; ESC C (Cursor Right)
		rjmp lcd_esc_perform_left			; ESC D (Cursor Left)
		rjmp lcd_esc_perform_clear			; ESC E (Clear Home)
		rjmp lcd_esc_perform_graphics_on	; ESC F (Enter graphics mode) 
		rjmp lcd_esc_perform_graphics_off	; ESC G (Exit graphics mode)
		rjmp lcd_esc_perform_home			; ESC H (Cursor Home)
		rjmp lcd_esc_perform_move_up		; ESC I (Cursor Up)
		rjmp lcd_esc_perform_clear_end		; ESC J (Clear Remaining Screen) 
		rjmp lcd_esc_perform_clear_line_end	; ESC K (Delete Rest of Line)
		rjmp lcd_esc_perform_line_insert	; ESC L (Insert Line)
		rjmp lcd_esc_perform_line_delete	; ESC M (Delete Line)
		rjmp lcd_esc_perform_pos			; ESC Y (Positioning Cursor)
		rjmp lcd_esc_perform_set_fg			; ESC b (Select Font Color)
		rjmp lcd_esc_perform_set_bg			; ESC c (Select Background Color)
		rjmp lcd_esc_perform_clear_start	; ESC d (Clear Screen up to Cursor Position)
		rjmp lcd_esc_perform_cur_enable		; ESC e (Cursor On)
		rjmp lcd_esc_perform_cur_disable	; ESC f (Cursor Off)
		rjmp lcd_esc_perform_cur_backup		; ESC j (Save Cursor Position)
		rjmp lcd_esc_perform_cur_restore	; ESC k (Restore Cursor to Saved Position)
		rjmp lcd_esc_perform_clear_line		; ESC l (Delete Line)
		rjmp lcd_esc_perform_clear_line_st	; ESC o (Delete Beginning of Line)
		rjmp lcd_esc_perform_reverse_on		; ESC p (Reverse on)	
		rjmp lcd_esc_perform_reverse_off	; ESC q (Reverse off)
		rjmp lcd_esc_perform_newline_on		; ESC v (Automatic Overflow On)
		rjmp lcd_esc_perform_newline_off	; ESC w (Automatic Overflow off)

; ### ESC A (Cursor Up)
lcd_esc_perform_up:							
		cpse cury,zero
		dec cury
		clr blink
		mcall lcd_esc_reset
		ret

; ### ESC B (Cursor Down)
lcd_esc_perform_down:					
		push a

		mov a,cury
		inc a
		cpi a,LCD_LINES
		breq lcd_esc_perform_down_done
		mov cury,a
lcd_esc_perform_down_done:
		clr blink
		mcall lcd_esc_reset

		pop a
		ret

; ### ESC C (Cursor Right)
lcd_esc_perform_right:				
		push a
	
		mov a,curx
		inc a
		cpi a,LCD_COLS
		breq lcd_esc_perform_right_done
		mov curx,a
lcd_esc_perform_right_done:
		clr blink
		mcall lcd_esc_reset

		pop a
		ret

; ### ESC D (Cursor Left)
lcd_esc_perform_left:						
		cpse curx,zero
		dec curx
lcd_esc_perform_left_done:
		clr blink
		mcall lcd_esc_reset
		ret

; ### ESC E (Clear Home)
lcd_esc_perform_clear:						
		mcall lcd_clear
		mcall lcd_esc_reset
		ret

; ### ESC F (Enter graphics mode)
lcd_esc_perform_graphics_on:  
		push a
		
		ser a
		mov gr,a
		mcall lcd_esc_reset

		pop a
		ret

; ### ESC G (Exit graphics mode)
lcd_esc_perform_graphics_off:
		clr gr
		mcall lcd_esc_reset
		ret

; ### ESC H (Cursor Home)	
lcd_esc_perform_home:					
		mcall lcd_home
		mcall lcd_esc_reset
		ret

; ### ESC I (Cursor Up)	
lcd_esc_perform_move_up:
		mcall lcd_move_up
		mcall lcd_esc_reset
		ret

; ### ESC J (Clear Remaining Screen) 
lcd_esc_perform_clear_end:
		mcall lcd_clear_end
		mcall lcd_esc_reset
		ret

; ### ESC K (Delete Rest of Line)
lcd_esc_perform_clear_line_end:
		mcall lcd_clear_line_end
		mcall lcd_esc_reset
		ret

; ### ESC L (Insert Line)
lcd_esc_perform_line_insert:
		mcall lcd_line_insert
		mcall lcd_esc_reset
		ret

; ### ESC M (Delete Line)
lcd_esc_perform_line_delete:
		mcall lcd_line_delete
		mcall lcd_esc_reset
		ret

; ### ESC Y (Positioning Cursor)
lcd_esc_perform_pos:						
		push a
		push b
		push c

		lds a,LCD_ESC_LEN
		cpi a,4
		brlo lcd_esc_perform_pos_exit
		lds c,LCD_ESC_BUF+2
		subi c,0x20
		lds b,LCD_ESC_BUF+3
		subi b,0x20
		mcall lcd_cursor_set
		mcall lcd_esc_reset

lcd_esc_perform_pos_exit:
		pop c
		pop b
		pop a
		ret

; ### ESC b (Select Font Color)
lcd_esc_perform_set_fg:
		push a
		
		lds a,LCD_ESC_LEN
		cpi a,3
		brlo lcd_esc_perform_set_fg_exit

		clr fg
		lds a,LCD_ESC_BUF+2
		andi a,0x0F
		cpse a,zero
		com fg
		mcall lcd_esc_reset

lcd_esc_perform_set_fg_exit:
		pop a
		ret

; ### ESC c (Select Background Color)
lcd_esc_perform_set_bg:
		push a

		lds a,LCD_ESC_LEN
		cpi a,3
		brlo lcd_esc_perform_set_bg_exit

		clr bg
		lds a,LCD_ESC_BUF+2
		andi a,0x0F
		cpse a,zero
		com bg
		mcall lcd_esc_reset

lcd_esc_perform_set_bg_exit:
		pop a
		ret

; ### ESC d (Clear Screen up to Cursor Position)
lcd_esc_perform_clear_start:
		mcall lcd_clear_start
		mcall lcd_esc_reset
		ret

; ### ESC e (Cursor On)
lcd_esc_perform_cur_enable:
		mcall lcd_cursor_enable
		mcall lcd_esc_reset
		ret

; ### ESC f (Cursor Off)
lcd_esc_perform_cur_disable:
		mcall lcd_cursor_disable
		mcall lcd_esc_reset
		ret

; ### ESC j (Save Cursor Position)
lcd_esc_perform_cur_backup:
		mov curbx,curx						; Backup Cursor Position
		mov curby,cury
		mcall lcd_esc_reset
		ret

; ### ESC k (Restore Cursor to Saved Position)
lcd_esc_perform_cur_restore:
		push b
		push c

		mov b,curbx							; Set Cursor-Position from Backup
		mov c,curby
		mcall lcd_cursor_set
		mcall lcd_esc_reset

		pop c
		pop b
		ret

; ### ESC l (Delete Line)
lcd_esc_perform_clear_line:
		mcall lcd_clear_line
		mcall lcd_esc_reset
		ret

; ### ESC o (Delete Beginning of Line)
lcd_esc_perform_clear_line_st:
		mcall lcd_clear_line_start
		mcall lcd_esc_reset
		ret

; ### ESC p (Reverse on)	
lcd_esc_perform_reverse_on:	
		push a
		
		ser a
		mov bg,a
		clr fg
		mcall lcd_esc_reset

		pop a
		ret

; ### ESC q (Reverse off)	
lcd_esc_perform_reverse_off:
		push a

		ser a
		mov fg,a
		clr bg		
		mcall lcd_esc_reset

		pop a
		ret

; ## ESC v (Automatic Overflow On)
lcd_esc_perform_newline_on:
		push a
		
		ser a
		mov nl,a
		mcall lcd_esc_reset

		pop a
		ret

; ## ESC w (Automatic Overflow off)
lcd_esc_perform_newline_off:
		clr nl
		mcall lcd_esc_reset
		ret

; *****************************************************************************
; *** SPI (Serial Serial Peripheral Interface)                              ***
; *****************************************************************************

; ### Init SPI
spi_init:
		push a
		push pdat

		in a,DDRB							; Prepare pins
		cbr a,PINB_CSN|PINB_MOSI|PINB_SCK
		sbr a,PINB_MISO|PINB_BUSY
		out DDRB,a
		in pdat,PORTB
		sbr pdat,PINB_CSN|PINB_MOSI|PINB_SCK
		cbr pdat,PINB_MISO|PINB_BUSY
		out PORTB,pdat	
		
		clr a								; Configure SPI
		sbr a,(1<<SPIE)						; SPIE (SPI interrupt enable)
		sbr a,(1<<SPE)						; SPE (SPI enable)
		sbr a,(0<<DORD)						; DORD (Data order)				-> 0: MSB of the data word is transmitted first / 1: LSB of the data word is transmitted first.
		sbr a,(0<<MSTR)						; MSTR (Master/slave select)	-> 0: Slave / 1: Master 
		sbr a,(0<<CPOL)						; CPOL (Clock polarity)			-> 0: Leading edge = Rising + Trailing edge = Falling (SCK low when idle) / 1: Leading edge = Falling + Trailing edge = Rising
		sbr a,(0<<CPHA)						; CPHA (Clock phase)			-> 0: Leading edge = Sample + Trailing edge = Setup / 1: Leading edge = Setup + Trailing edge = Sample
		sbr a,(0<<SPR0)|(0<<SPR1)			; SPRx (SPI clock rate select)	-> 000: fosc/4, 001: fosc/16, 010: fosc/64, 011: fosc/128, 100: fosc/2, 101: fosc/8, 110: fosc/32, 111: fosc/64
		out SPCR,a

		clr a
		sbr a,(1<<SPI2X)					; SPI2X: Enable Double SPI speed bit (see SPRx)
		out SPSR,a

		pop pdat
		pop a
		ret

; ### Receive byte from SPI (interrupt)
spi_recv_intr:
		push a
		in statreg,SREG						

		in a,PORTB							; Set BUSY Pin
		sbr a,PINB_BUSY
		out PORTB,a

		in a,PORTD							; Disable Display
		cbr a,PIND_DISP
		out PORTD,a

		mcall lcd_esc_get_len				; Check For Escape Sequence
		tst a
		breq spi_recv_int_check
		in a,SPDR							; Add Char from SPI to Escape Sequence
		mcall lcd_esc_add
		mcall lcd_esc_perform
		rjmp spi_recv_intr_done

spi_recv_int_check:
		in a,SPDR							; Get Char from SPI 
		cpi a,0x7F							; >$7F: DEL (Delete)?
		brne spi_recv_int_bs
		mcall lcd_delete
		rjmp spi_recv_intr_done				 
spi_recv_int_bs:							; >$08: BS (Back Space)?
		cpi a,0x08
		brne spi_recv_int_nl
		mcall lcd_delete 
		rjmp spi_recv_intr_done
spi_recv_int_nl:							; >$0A: LF (Line Feed)?
		cpi a,'\n'
		brne spi_recv_int_ff		
		mcall lcd_new_line
		rjmp spi_recv_intr_done
spi_recv_int_ff:							; >$0C: FF (Form Feed)?
		cpi a,0x0C
		brne spi_recv_int_cr
		mcall lcd_clear
		rjmp spi_recv_intr_done
spi_recv_int_cr:							; >$0D: CR (Carrige Return)?
		cpi a,'\r'
		brne spi_recv_int_cur_enable
		mcall lcd_start
		rjmp spi_recv_intr_done
spi_recv_int_cur_enable:					; >$0E: SO (Enable Cursor) ?
		cpi a,0x0E
		brne spi_recv_int_cur_disable
		mcall lcd_cursor_enable
		rjmp spi_recv_intr_done
spi_recv_int_cur_disable:					; >$0F: SI (Disblae Cursor) ?
		cpi a,0x0F
		brne spi_recv_int_esc
		mcall lcd_cursor_disable
		rjmp spi_recv_intr_done
spi_recv_int_esc:
		cpi a,0x1B							; >$1B: ESC (Escape Sequence)?
		brne spi_recv_intr_char
		mcall lcd_esc_add
		rjmp spi_recv_intr_done
spi_recv_intr_char:							; ASCII Char?
		cpi a,' '
		brlo spi_recv_intr_done
		cpi a,0x80
		brsh spi_recv_intr_done
		mcall lcd_add

spi_recv_intr_done:
		in a,PORTD							; Enable Display
		sbr a,PIND_DISP
		out PORTD,a

		in a,PORTB							; Reset BUSY Pin
		cbr a,PINB_BUSY
		out PORTB,a

spi_recv_intr_exit:
		out SREG,statreg					
		pop a
		reti

; *****************************************************************************
; *** Utility                                                               ***
; *****************************************************************************

; ### Converts Byte to Decimal (2 Digits in ASCII)
; >a: Byte
; <b: Digit #1 
; <c: Digit #2
util_byte2dec2:
		push a
	    
		ldi b,'0'-1							; Digit #1 
util_byte2dec2_1:
		inc b 
		subi a,10
		brcc util_byte2dec2_1 
		subi a,-10  
		ldi c,'0'							; Digit #2
		add c,a
		   
		pop a
		ret 

; ### Converts Byte to Decimal (2 Digits in Status Line)
; >a: Byte
; <b: Digit #1 
; <c: Digit #2
util_byte2dec_s:
		push a
	    
		ldi b,'0'+0x60-1						; Digit #1 
util_byte2dec_s_1:
		inc b 
		subi a,10
		brcc util_byte2dec_s_1 
		subi a,-10  
		ldi c,'0'+0x60							; Digit #2
		add c,a
		   
		pop a
		ret 
