; Z80V1 (CP/M V1)
; (C) Mario Latzig 2015
;
; on ATmega32
;
; PA0: D0	  (Z80 CPU)
; PA1: D1	  (Z80 CPU)
; PA2: D2	  (Z80 CPU)
; PA3: D3	  (Z80 CPU)
; PA4: D4	  (Z80 CPU)
; PA5: D5	  (Z80 CPU)
; PA6: D6	  (Z80 CPU)
; PA7: D7	  (Z80 CPU)
;
; PB0: SCL    (I2C)
; PB1: SDA    (I2C)
; PB2: HALT   (Z80 CPU, In)
; PB3: LED	  (External LED)
; PB4: CE     (HW SPI: RAM/IDE) 
; PB5: MOSI   (HW SPI: RAM/IDE)
; PB6: MISO   (HW SPI: RAM/IDE)
; PB7: SCK    (HW SPI: RAM/IDE)
;
; PC0: CCS    (Console LCD Chip-Select)
; PC1: CBUSY  (Console LCD Busy)
; PC2: KCS	  (Console Keyboard Chip-Select)
; PC3: KONOFF (Console Keyboard "ON/OFF" Key Pressed -> Reset)
;
; PD0: RXD    (UART/RS232)
; PD1: TXD    (UART/RS232)
; PD2: IOREQ  (Z80 CPU, In)
; PD3: MREQ   (Z80 CPU, In)
; PD4: RD     (Z80 CPU, In)
; PD5: WR     (Z80 CPU, In)    
; PD6: WAIT   (Z80 CPU, Output) 
; PD7: RESET  (Z80 CPU, Output)

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************
.equ FC = 18432000							; Clock = 18.432 MHz

; ### PINS ####################################################################
.equ PINB_SCL = (1<<0)						; PB0: SCL    (I2C)
.equ PINB_SDA = (1<<1)						; PB1: SDA    (I2C)
.equ PINB_HALT = (1<<2)						; PB2: HALT   (Z80 CPU, In)
.equ PINB_LED = (1<<3)						; PB3: LED	  (External LED)
.equ PINB_CSN = (1<<4)						; PB4: CE     (HW SPI: RAM/IDE) 
.equ PINB_MOSI = (1<<5)						; PB5: MOSI   (HW SPI: RAM/IDE)
.equ PINB_MISO = (1<<6)						; PB6: MISO   (HW SPI: RAM/IDE)
.equ PINB_SCK = (1<<7)						; PB7: SCK    (HW SPI: RAM/IDE)

.equ PINC_CCS = (1<<0)						; PC0: CCS    (Console LCD Chip-Select)
.equ PINC_CBUSY = (1<<1)					; PC1: CBUSY  (Console LCD Busy)
.equ PINC_KCS = (1<<2)						; PC2: KCS    (Console Keyboard Chip-Select)
.equ PINC_KONOFF = (1<<3)					; PC3: KONOFF (Console Keyboard "ON/OFF" Key Pressed -> Reset)

.equ PIND_RXD = (1<<0)						; PD0: RXD    (UART/RS232)
.equ PIND_TXD = (1<<1)						; PD1: TXD    (UART/RS232)
.equ PIND_IOREQ = (1<<2)					; PD2: IOREQ  (Z80 CPU, In)
.equ PIND_MREQ = (1<<3)						; PD3: MREQ   (Z80 CPU, In)
.equ PIND_RD = (1<<4)						; PD4: RD     (Z80 CPU, In)
.equ PIND_WR = (1<<5)						; PD5: WR     (Z80 CPU, In)    
.equ PIND_WAIT = (1<<6)						; PD6: WAIT   (Z80 CPU, Out) 
.equ PIND_RESET = (1<<7)					; PD7: RESET  (Z80 CPU, Out)

; === Bus =====================================================================
.equ BUS_IO_CONSOLE = 0x00					; IO $00: Console In & Out
.equ BUS_IO_CONSTAT = 0x01					; IO $01: Console Status
.equ BUS_IO_SERIAL = 0x02					; IO $02: Serial In & Out
.equ BUS_IO_SERSTAT = 0x03					; IO $03: Serial Status
.equ BUS_IO_DISK_IO	= 0x08					; IO $08: Disk I/O (Read/Write)
.equ BUS_IO_DISK_SELDSK	= 0x09				; IO $09: Disk Number
.equ BUS_IO_DISK_SETTRK_HI = 0x0A			; IO $0A: Disk Track (HI Byte)
.equ BUS_IO_DISK_SETTRK_LO = 0x0B			; IO $0B: Disk Track (LO Byte)
.equ BUS_IO_DISK_SETSEC_HI = 0x0C			; IO $0C: Disk Sector (HI Byte)
.equ BUS_IO_DISK_SETSEC_LO = 0x0D			; IO $0D: Disk Sector (LO Byte)
.equ BUS_IO_DISK_STATUS	= 0x0E				; IO $0E: Disk Status (Read-Only)
.equ BUS_IO_BUFFER = 0x0F					; IO $0F: Buffer In & Out
.equ BUS_IO_ADDR = 0x20						; IO $20-$9F: Buffer Address (128 Bytes)
.equ BUS_IO_NOP = 0xFF						; IO $FF: (No Operation)
.equ BUS_IO_NUM = 16						; Number of IO Operations ($00-$0F)

.equ BUS_BUF_SIZE = 128						; Size of IO Buffer

; === Disk ====================================================================
.equ DISK_NUM_MAX = 16						; Maximal Number of Disks
.equ DISK_NUM_TRKS = 2000					; Number of Tracks
.equ DISK_NUM_TRKSECS = 32					; Number of Sectors per Track
.equ DISK_NUM_SECS = DISK_NUM_TRKS*DISK_NUM_TRKSECS
.equ DISK_NUM_LBA = DISK_NUM_SECS / 4		; Num of LBAs per Disk

; === Monitor =================================================================
.equ MONITOR_START = 0xD000					; Start Address of "Z80 Monitor"

; === CPM =====================================================================
.equ CPM_BIOS_START = 0xFA00				; Start Address of CP/M Bios (Cold Start)
.equ CPM_BOOT_CPM = 0xE000					; Start Address of CP/M Boot Range ($E000-$FFFF = 8 KB = 64 Sectors)
.equ CPM_BOOT_SECS = 64						; Number of Boot Sectors 
.equ CPM_BOOT_START_TRK = 0					; Start Track for Boot
.equ CPM_BOOT_START_SEC = 0					; Start Sector for Boot

.eseg
; *****************************************************************************
; *** EEPROM                                                                ***
; *****************************************************************************

MAIN_EETXT_INIT_CON:						; Inits Console (Escape Sequence)
	.db 27,'E',27,'G',27,'e',27,'q',27,'v',0
MAIN_EETXT_TITLE:							; Title
	.db "Z80 SBC (C) Mario Latzig 2015",0	
MAIN_EETXT_OK:								; "OK"
	.db "OK",0
MAIN_EETXT_ERROR_RAM:						; "CRC Error!"
	.db "CRC Error at $",STR_FMTID_HEX,STR_FMTID_HEX,0
MAIN_EETXT_CHECK_RAM:						; "Check RAM"
	.db "- Check RAM (64KB): ",0
MAIN_EETXT_ERROR_IDE:						; "IDE Error!"
	.db "Error $",STR_FMTID_HEX,0
MAIN_EETXT_CHECK_IDE:						; "Check IDE"
	.db "- Check IDE: ",0
MAIN_EETXT_BOOT:							; "Boot CP/M"
	.db "Booting CP/M, Press ESC to start Monitor",0
MAIN_EETXT_BOOT_ERROR:						; "Boot CP/M"
	.db "Boot Error $",STR_FMTID_HEX,0

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

; === Bus =====================================================================
BUS_IO_CMD:		.byte 1						; Operation of IO (see BUS_IO_xxx)
BUS_IO_DATA:	.byte 1						; Data Byte of IO
BUS_IO_DIR:		.byte 1						; Flag of IO Direction (0=Read,>0:Write)
BUS_RAM_CRC:	.byte 2						; 2-Byte Buffer for RAM CRC
BUS_BUF:		.byte BUS_BUF_SIZE			; IO Buffer (128 Bytes)

; === Disk ====================================================================
DISK_NUM:		.byte 1						; Current Disk (0-15)
DISK_TRACK_HI:	.byte 1						; Current Track (HI byte)
DISK_TRACK_LO:	.byte 1						; Current Track (LO byte)
DISK_SECTOR_HI:	.byte 1						; Current Sector (HI byte)
DISK_SECTOR_LO:	.byte 1						; Current Sector (LO byte)
DISK_STATUS:	.byte 1						; Current Status Flag (0=OK,>0:ERROR)

; === Console =================================================================
CON_KEY:		.byte 1						; Key Code from Console
CON_KEY_MAP:	.byte 1						; Index for Key Mapping

.cseg
; *****************************************************************************
; *** Vectors                                                               ***
; *****************************************************************************

.org 0x0000
		jmp init
.org INT0addr								; Interuupt for INT0 (IOREQ)
		jmp bus_isr_int0					
.org INT_VECTORS_SIZE

; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

; === Include "Z80 Monitor" ===================================================
.include "monitor.inc"
MONITOR_END:
.equ MONITOR_SIZE = 2*(MONITOR_END-MONITOR)

; === Console =================================================================
CON_KEY_MAP_SETUP:          				; Console Key Mapping (Control Based)
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $00-$0F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $10-$1F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $20-$2F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $30-$3F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $40-$4F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $50-$5F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $60-$6F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $70-$7F

	.db 0   ,0   ,0   ,0   ,'E' ,'X' ,'S' ,'D' ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $80-$8F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $90-$9F

CON_KEY_MAP_FONT:          					; Console Key Mapping (Key-Based)
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $00-$0F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $10-$1F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $20-$2F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $30-$3F

	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $40-$4F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $50-$5F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $60-$6F
	.db 0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $70-$7F

	.db 0   ,0   ,0   ,0   ,0x05,0x18,0x13,0x04,0   ,0   ,0   ,0   ,0x12,0x03,0   ,0	; $80-$8F
	.db 0   ,0x01,0x06,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0   ,0	; $90-$9F

; *****************************************************************************
; *** Main                                                                  ***
; *****************************************************************************

; === Include Modules =========================================================
.include "util.inc"
.equ STR_BUF_SIZE = 80					
.equ STR_FMT_SIZE = 16
.include "str.inc"
.include "math.inc"
.include "crc8.inc"
#define UART_NO_IRQ
.equ UART_BAUD = 9600
.include "uart.inc"
#define MCP23S17_PINS
.equ MCP23S17_PIN_CSN = PINB_CSN
.equ MCP23S17_PIN_MOSI = PINB_MOSI		
.equ MCP23S17_PIN_MISO = PINB_MISO		
.equ MCP23S17_PIN_SCK = PINB_SCK		
.equ MCP23S17_PORTDAT = PORTB
.equ MCP23S17_PORTDIR = DDRB
#define MCP23S17_SPI_SPEED
.equ MCP23S17_SPR1 = (0<<SPR0)|(0<<SPR1)
.equ MCP23S17_SPR2 = (0<<SPI2X)
.include "mcp23s17.inc"
#define IDE_SPI
.equ IDE_GPIO_ADDR2 = MCP23S17_ADDR4
.equ IDE_GPIO_ADDR = MCP23S17_ADDR5
.include "ide.inc"

; ### INIT ####################################################################
init:
		cli									; Disable interrupts
		
		ldi a,LOW(RAMEND)					; Init stack pointer
		out SPL,a
		ldi a,HIGH(RAMEND)
		out SPH,a

		clr init0							; Reset init0 register
		mcall util_init						; Init Standard Modules
		mcall str_init
		mcall math_init
		mcall crc8_init
		mcall uart_init
		mcall mcp23s17_init
		mcall ide_init
		
		mcall ser_init						; Inits Internal Modules
		mcall con_init
		mcall bus_init
		mcall disk_init

		sei									

; ### MAIN ####################################################################
main:
		ldi a,100
		mcall util_delay_Xms

		ldi a,CHR_FF						; Clear Console & Serial
		mcall con_put
		mcall ser_put
		mcall str_clear
		ldi zl,LOW(MAIN_EETXT_INIT_CON)		; Init Console (Escape Sequence)
		ldi zh,HIGH(MAIN_EETXT_INIT_CON)
		mcall str_add_str
		mcall con_str
		mcall str_clear						; Show Title (Console & Serial)
		ldi zl,LOW(MAIN_EETXT_TITLE)		
		ldi zh,HIGH(MAIN_EETXT_TITLE)
		mcall str_add_str
		mcall con_str
		mcall con_nl
		mcall ser_str
		mcall ser_nl

main_ram_init:								; ### Initialize RAM
		mcall str_clear						; Show "Check RAM" Message
		ldi zl,LOW(MAIN_EETXT_CHECK_RAM)	
		ldi zh,HIGH(MAIN_EETXT_CHECK_RAM)
		mcall str_add_str
		mcall con_str
		mcall bus_ram_start
main_ram_init_crc:							; Fill RAM with CRC Bytes
		clr zl
		clr zh
main_ram_init_crc_loop:						
		sts BUS_RAM_CRC+0,zl
		sts BUS_RAM_CRC+1,zh
		ldi a,2
		push zl
		push zh
		ldi zl,LOW(BUS_RAM_CRC)
		ldi zh,HIGH(BUS_RAM_CRC)
		mcall crc8_calc
		pop zh
		pop zl
		mcall bus_ram_set
		adiw z,1
		brne main_ram_init_crc_loop
main_ram_init_chk:							; Check RAM with CRC Bytes
		clr zl
		clr zh
main_ram_init_chk_loop:						
		sts BUS_RAM_CRC+0,zl
		sts BUS_RAM_CRC+1,zh
		ldi a,2
		push zl
		push zh
		ldi zl,LOW(BUS_RAM_CRC)
		ldi zh,HIGH(BUS_RAM_CRC)
		mcall crc8_calc
		pop zh
		pop zl
		mov b,a
		mcall bus_ram_get					; Read RAM Byte and Check against CRC Byte 
		cp a,b
		breq main_ram_init_chk_cont
main_ram_init_chk_error:					; => Show "Error" and stop!
		mcall str_clear						
		mcall str_fmt_reset
		mov a,zh
		mcall str_fmt_add
		mov a,zl
		mcall str_fmt_add
		ldi zl,LOW(MAIN_EETXT_ERROR_RAM)	
		ldi zh,HIGH(MAIN_EETXT_ERROR_RAM)
		mcall str_add_fmt
		mcall con_str
main_ram_init_chk_error_stop:
		rjmp main_ram_init_chk_error_stop							
main_ram_init_chk_cont:						
		clr a								; => CRC OK, Reset RAM Byte
		mcall bus_ram_set
		adiw z,1
		brne main_ram_init_chk_loop
		mcall str_clear						; Finish, Show "OK" 
		ldi zl,LOW(MAIN_EETXT_OK)	
		ldi zh,HIGH(MAIN_EETXT_OK)
		mcall str_add_fmt
		mcall con_str
		mcall con_nl

main_ide_init:								; ### Initialize IDE
		mcall str_clear						; Show "Check IDE" Message
		ldi zl,LOW(MAIN_EETXT_CHECK_IDE)	
		ldi zh,HIGH(MAIN_EETXT_CHECK_IDE)
		mcall str_add_fmt
		mcall con_str

		mcall _ide_check_error				; Check Error Code for IDE (should be Zero)
		tst pdat
		breq main_ide_init_ok
main_ide_init_chk_error:					; => Show "Error" and stop!
		mcall str_clear						
		mcall str_fmt_reset
		mov a,pdat
		mcall str_fmt_add
		ldi zl,LOW(MAIN_EETXT_ERROR_IDE)	
		ldi zh,HIGH(MAIN_EETXT_ERROR_IDE)
		mcall str_add_fmt
		mcall con_str
main_ide_init_chk_error_stop:
		rjmp main_ide_init_chk_error_stop							

main_ide_init_ok:							; => IDE OK
		mcall str_clear						; Show "OK" 
		ldi zl,LOW(MAIN_EETXT_OK)	
		ldi zh,HIGH(MAIN_EETXT_OK)
		mcall str_add_fmt
		mcall con_str
		mcall con_nl
		
main_ide_init_identify:						; Perform IDENTIFY Command on IDE
		ldi pdat,0							
		mcall ide_identify
main_ide_init_serial:
		ldi a,'-'
		mcall con_put
		ldi a,' '
		mcall con_put
		mcall con_put
		mcall con_put
		ldi ol,LOW(IDE_ID_SERIAL_NUMBER)	; Prints Serial Number
		ldi oh,HIGH(IDE_ID_SERIAL_NUMBER)
		ldi zl,LOW(IDE_ID_CONTROLLER_TYPE)
		ldi zh,HIGH(IDE_ID_CONTROLLER_TYPE)
main_ide_init_serial_loop:
		mcall ide_buf_get
		cpi a,' '
		brlo main_ide_init_serial_cont
		cpi a,0x7F
		brsh main_ide_init_serial_cont
		mcall con_put
main_ide_init_serial_cont:
		adiw ol:oh,1
		cp ol,zl
		cpc oh,zh
		brne main_ide_init_serial_loop		
		mcall con_nl

main_ide_init_firmware:						; Prints Firmware Number
		ldi a,'-'
		mcall con_put
		ldi a,' '
		mcall con_put
		mcall con_put
		mcall con_put
		ldi ol,LOW(IDE_ID_CONTROLLER_FIRMWARE)	
		ldi oh,HIGH(IDE_ID_CONTROLLER_FIRMWARE)
		ldi zl,LOW(IDE_ID_MODEL_NUMBER)
		ldi zh,HIGH(IDE_ID_MODEL_NUMBER)
main_ide_init_firmware_loop:
		mcall ide_buf_get
		cpi a,' '
		brlo main_ide_init_firmware_cont
		cpi a,0x7F
		brsh main_ide_init_firmware_cont
		mcall con_put
main_ide_init_firmware_cont:
		adiw ol:oh,1
		cp ol,zl
		cpc oh,zh
		brne main_ide_init_firmware_loop		
		mcall con_nl

main_ide_init_model:						; Prints Model Number
		ldi a,'-'
		mcall con_put
		ldi a,' '
		mcall con_put
		mcall con_put
		mcall con_put
		ldi ol,LOW(IDE_ID_MODEL_NUMBER)	
		ldi oh,HIGH(IDE_ID_MODEL_NUMBER)
		ldi zl,LOW(IDE_ID_SECTORS_INTERRUPT)
		ldi zh,HIGH(IDE_ID_SECTORS_INTERRUPT)
main_ide_init_model_loop:
		mcall ide_buf_get
		cpi a,' '
		brlo main_ide_init_model_cont
		cpi a,0x7F
		brsh main_ide_init_model_cont
		mcall con_put
main_ide_init_model_cont:
		adiw ol:oh,1
		cp ol,zl
		cpc oh,zh
		brne main_ide_init_model_loop		
main_ide_init_done:
		mcall con_nl

main_boot:									; ### Boots CP/M
		mcall str_clear						; Show "Boot CP/M" (on Console & Serial)
		ldi zl,LOW(MAIN_EETXT_BOOT)	
		ldi zh,HIGH(MAIN_EETXT_BOOT)
		mcall str_add_fmt
		mcall con_str
		mcall con_nl
		mcall ser_str
		mcall ser_nl
		ser a								; Load CP/M with Progress
		mcall cpm_load
main_boot_done:								; => Boot Done
		cpse a,zero
		rjmp main_boot_start_monitor
main_boot_start_cpm:						; ### Starts CP/M
		mcall con_nl
		mcall ser_nl
		mcall cpm_reset
		rjmp main_loop
main_boot_start_monitor:					; ### Starts "Z80 Monitor"
		mcall monitor_run	
		rjmp main_loop
main_boot_error:							; => Show "Error" and stop!
		mcall con_nl
		mcall ser_nl
		mcall str_clear						
		mcall str_fmt_reset
		mcall str_fmt_add
		ldi zl,LOW(MAIN_EETXT_BOOT_ERROR)	
		ldi zh,HIGH(MAIN_EETXT_BOOT_ERROR)
		mcall str_add_fmt
		mcall con_str
		mcall ser_str
main_boot_error_stop:
		rjmp main_boot_error_stop							

main_loop:									; ### Main Loop (Running Z80)
main_loop_chk_reset:						; > Check for Reset (User pressed "ON/OFF" Key)
		mcall con_is_reset
		tst a
		breq main_loop_chk_halt
main_loop_reset:							; => RESET
		mcall con_is_reset					; Wait
		cpse a,zero
		rjmp main_loop_reset				
		mcall cpm_load						; Load CP/M for Cold Start
		mcall cpm_reset					
		rjmp main_loop

main_loop_chk_halt:							; > Check For HALT
		in b,PINB
		andi b,PINB_HALT
		cpse b,zero
		rjmp main_loop
main_halt:									; => HALT 
		mcall cpm_load						; Load CP/M for Cold Start
		mcall monitor_run					; Try to run Monitor!
		jmp main_loop

; ### Load CP/M from Disk
; <a: The Flag for ESC Pressed
cpm_load:
		push b
		push c
		push d
		push e
		push xl
		push xh
		push zl
		push zh

		cli

		mov e,a								; (e = Flag for Showing Progess)

		clr a								; Reset Disk (Reset to Track/Sector for Bios of Disk A)
		mcall disk_num_set 
		ldi a,HIGH(CPM_BOOT_START_TRK)
		mcall disk_track_set_hi
		ldi a,LOW(CPM_BOOT_START_TRK)
		mcall disk_track_set_lo
		ldi a,HIGH(CPM_BOOT_START_SEC)
		mcall disk_sector_set_hi
		ldi a,LOW(CPM_BOOT_START_SEC)
		mcall disk_sector_set_lo
		ldi b,CPM_BOOT_SECS
		clr c								; (c = Flag for Monitor, ESC pressed)
		ldi zl,LOW(CPM_BOOT_CPM)
		ldi zh,HIGH(CPM_BOOT_CPM)
		mcall bus_ram_start

cpm_load_loop:								; Read Sector
		tst e
		breq cpm_load_loop_read
		ldi a,'.'
		mcall con_put
		mcall ser_put
		ldi a,25
		mcall util_delay_Xms
cpm_load_loop_read:
		mcall disk_read
		mcall disk_get_status					
		tst a
		brne cpm_load_error
cpm_load_copy:								; Copy Sector to Memory
		ldi xl,LOW(BUS_BUF)
		ldi xh,HIGH(BUS_BUF)
		ldi d,BUS_BUF_SIZE
cpm_load_copy_loop:
		ld a,x+
		mcall bus_ram_set
		adiw z,1
		dec d
		brne cpm_load_copy_loop
		mcall con_get
		cpi a,27							; ESCP pressed?
		breq cpm_load_loop_mark_monitor
		mcall ser_get
		cpi a,27
		brne cpm_load_loop_next
cpm_load_loop_mark_monitor:
		ser c
cpm_load_loop_next:						; Get Next Sector
		mcall disk_sector_get_lo
		inc a
		cpi a,DISK_NUM_TRKSECS
		brne cpm_load_loop_next_cont
		mcall disk_track_get_lo
		inc a
		mcall disk_track_set_lo
		clr a
cpm_load_loop_next_cont:
		mcall disk_sector_set_lo		
		dec b
		brne cpm_load_loop
		mov a,c
		rjmp cpm_load_exit

cpm_load_error:								; => Show "Error" and stop!
		mcall con_nl
		mcall ser_nl
		mcall str_clear						
		mcall str_fmt_reset
		mcall str_fmt_add
		ldi zl,LOW(MAIN_EETXT_BOOT_ERROR)	
		ldi zh,HIGH(MAIN_EETXT_BOOT_ERROR)
		mcall str_add_fmt
		mcall con_str
		mcall ser_str
cpm_load_error_stop:
		rjmp cpm_load_error_stop							

cpm_load_exit:
		sei

		pop zh
		pop zl
		pop xh
		pop xl
		pop e
		pop d
		pop c
		pop b
		ret

; ### Reset Z80 for CP/M Cold Start
cpm_reset:
		push a
		push zl
		push zh
		
		cli

		mcall bus_ram_start					; Set Jump Instruction at $0000
		clr zl
		clr zh
		ldi a,0xC3							
		mcall bus_ram_set
		adiw z,1
		ldi a,LOW(CPM_BIOS_START)
		mcall bus_ram_set
		adiw z,1
		ldi a,HIGH(CPM_BIOS_START)
		mcall bus_ram_set
		mcall bus_ram_stop
		mcall bus_reset						; Reset Z80

		sei

		pop zh
		pop zl
		pop a
		ret

; *****************************************************************************
; *** Monitor ("Z80 Monitor")                                              ***
; *****************************************************************************
monitor_run:
		push a
		push ol
		push oh
		push xl
		push xh
		push zl
		push zh
		
		cli

		mcall bus_ram_start					
		ldi zl,LOW(MONITOR_START)
		ldi zh,HIGH(MONITOR_START)
		ldi xl,LOW(2*MONITOR)
		ldi xh,HIGH(2*MONITOR)
		ldi ol,LOW(MONITOR_SIZE)
		ldi oh,HIGH(MONITOR_SIZE)
monitor_copy:             					; Copy Binary to Memory
		push zl
		push zh
		mov zl,xl
		mov zh,xh
		lpm a,z+
		mov xl,zl
		mov xh,zh
		pop zh
		pop zl
		mcall bus_ram_set
		adiw z,1
		sbiw ol:oh,1
		brne monitor_copy
		clr zl
		clr zh
		ldi a,0xC3							; Set Jump Instruction at $0000
		mcall bus_ram_set
		adiw z,1
		ldi a,LOW(MONITOR_START)
		mcall bus_ram_set
		adiw z,1
		ldi a,HIGH(MONITOR_START)
		mcall bus_ram_set
		mcall bus_ram_stop
		mcall bus_reset						; Reset Z80

		sei

		pop zh
		pop zl
		pop xh
		pop xl
		pop oh
		pop ol
		pop a
		ret	

; *****************************************************************************
; *** Bus                                                                  ***
; *****************************************************************************

; ### Inits Bus
bus_init:
		push a
		push pdat

		in pdat,PORTD						; Set Pins for Bus Control (Inital Z80 Reset)
		sbr pdat,PIND_RD|PIND_WR|PIND_MREQ|PIND_IOREQ|PIND_WAIT						
		cbr pdat,PIND_RESET
		out PORTD,pdat
		in a,DDRD	
		cbr a,PIND_RD|PIND_WR|PIND_MREQ|PIND_IOREQ						
		sbr a,PIND_RESET|PIND_WAIT
		out DDRD,a
		in pdat,PORTB
		sbr pdat,PINB_HALT
		out PORTB,pdat
		in a,DDRB
		cbr a,PINB_HALT
		out DDRB,a

		ser a								; Enable Pull-Ups for Address Pins
		clr b
		mcall mcp23s17_set_pullup_a
		mcall mcp23s17_set_pullup_b

		in a,MCUCR							; Init Interrupt for IOREQ on Z80
		cbr a,(1<<ISC01)|(1<<ISC00)
		sbr a,(0<<ISC01)|(0<<ISC00)
		out MCUCR,a
		in a,GICR
		sbr a,(1<<INT0)
		out GICR,a

		ldi a,BUS_IO_NOP					; Reset Internal for IO Operations
		sts BUS_IO_CMD,a
		clr a
		sts BUS_IO_DATA,a
		sts BUS_IO_DIR,a

		pop pdat
		pop a
		ret

; ### Reset Z80
bus_reset:
		push a
		push pdat

		cli

		in pdat,PORTD						; Holds RESET for Z80 
		cbr pdat,PIND_RESET|PIND_WAIT
		out PORTD,pdat
		ldi a,100							; Delay 100ms
		mcall util_delay_Xms
		sbr pdat,PIND_RESET|PIND_WAIT
		out PORTD,pdat

		sei

		pop pdat
		pop a
		ret

; ### Start Access to RAM (Resets & Halt CPU)
bus_ram_start:
		push a
		push b
		push pdat

		clr a								; Set Address Pins to Output
		clr b
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_set_iodir_b

		in pdat,PORTD						; Perform Reset 
		cbr pdat,PIND_RESET|PIND_WAIT
		out PORTD,pdat
		ldi a,100							; Delay 100ms
		mcall util_delay_Xms
		sbr pdat,PIND_RD|PIND_WR			; Set RD & WR as Output
		cbr pdat,PIND_MREQ
		out PORTD,pdat
		in a,DDRD
		sbr a,PIND_RD|PIND_WR|PIND_MREQ
		out DDRD,a
	
		pop pdat
		pop b
		pop a
		ret

; ### Stop Access to RAM
bus_ram_stop:
		push a
		push b
		push pdat

		clr a								; Disable Data Pins (Input)
		out DDRA,a

		ser a								; Set Address Pins to Input
		clr b
		mcall mcp23s17_set_iodir_a
		mcall mcp23s17_set_iodir_b

		in pdat,PORTD						; Set RD & WR as Input						
		sbr pdat,PIND_RD|PIND_WR|PIND_MREQ
		out PORTD,pdat
		in a,DDRD	
		cbr a,PIND_RD|PIND_WR|PIND_MREQ
		out DDRD,a
		pop pdat
		pop b
		pop a
		ret

; ### Set RAM
; >a: The Byte 
; >z: The Address
bus_ram_set: 
		push a
		push b
		push pdat

		ser b								; Set Data Pins (Output)
		out DDRA,b
		out PORTA,a

		clr b								; Set Address Pins
		mov a,zl
		mcall mcp23s17_set_gpio_a
		mov a,zh
		mcall mcp23s17_set_gpio_b

		in pdat,PORTD						; Strobe WR
		cbr pdat,PIND_WR
		out PORTD,pdat
		nop									; (Delay, at least 70ns)
		nop
		sbr pdat,PIND_WR 
		out PORTD,pdat

		pop pdat
		pop b
		pop a
		ret

; ### Get RAM 
; <a: The Byte
; >z: The Address
bus_ram_get:
		push b
		push pdat		

		clr a								; Prepare Data Pins (Input)
		out DDRA,a

		clr b								; Set Address Pins
		mov a,zl
		mcall mcp23s17_set_gpio_a
		mov a,zh
		mcall mcp23s17_set_gpio_b

		in pdat,PORTD						; Strobe RD
		cbr pdat,PIND_RD
		out PORTD,pdat
		nop									; (Delay, at least 70ns)
		nop
		in a,PINA							; Get Data
		sbr pdat,PIND_RD
		out PORTD,pdat

		pop pdat
		pop b
		ret

; ### Interuupt for INT0 (IOREQ)
bus_isr_int0:
		push a
		push pdat
		in stat,SREG				

bus_isr_int0_dir:							; Detect Direction
		in pdat,PIND							
		andi pdat,PIND_RD
		tst pdat
		breq bus_isr_int0_dir_read
bus_isr_int0_dir_write:						; ==> Write to IO Port (Output)
		in pdat,PINA						; (Store Data Byte)
		sts BUS_IO_DATA,pdat
		ser a								; Preset Flag
		rjmp bus_isr_int0_dir_cont
bus_isr_int0_dir_read:						; ==> Read from IO Port (Input)
		clr pdat							; (Reset Data Byte, has to be set later)
		sts BUS_IO_DATA,pdat
		clr a								; Reset Flag
bus_isr_int0_dir_cont:
		sts BUS_IO_DIR,a					; (Store Direction)

bus_isr_int0_op:							; Detect New IO Operation (Address Pins)
		mcall mcp23s17_get_gpio_a			
		cpi a,BUS_IO_ADDR					; Translate Buffer Address, if needed...
		brlo bus_isr_int0_op_cont
		cpi a,BUS_IO_ADDR+BUS_BUF_SIZE
		brsh bus_isr_int0_exit
		subi a,BUS_IO_ADDR
		sts BUS_IO_CMD,a	
		ldi a,BUS_IO_BUFFER
		rjmp bus_isr_int0_op_perform
bus_isr_int0_op_cont:
		sts BUS_IO_CMD,a	
bus_isr_int0_op_perform:
		cpi a,BUS_IO_NUM
		brsh bus_isr_int0_exit
		ldi zl,LOW(bus_io_table)			; Perform IO Operation
		ldi zh,HIGH(bus_io_table)
		add zl,a
		adc zh,zero
		icall 

bus_isr_int0_read:							; Output Data Byte to IO Port
		lds a,BUS_IO_DIR
		tst a
		brne bus_isr_int0_unrelease
		ser a								; Set Data Pins
		out DDRA,a
		lds pdat,BUS_IO_DATA
		out PORTA,pdat

bus_isr_int0_unrelease:						; Unrelease WAIT
		in pdat,PORTD						
		cbr pdat,PIND_WAIT
		out PORTD,pdat
		nop									; (Delay for Reset WAIT)
		nop
		nop
		nop
		nop
		nop
		sbr pdat,PIND_WAIT
		out PORTD,pdat
		nop									; Inactivate Data Pins
		clr a								
		out DDRA,a

		ldi a,BUS_IO_NOP					; Reset IO Operation (No Operation)
		sts BUS_IO_CMD,a

bus_isr_int0_exit:
		out SREG,stat		
		pop pdat			
		pop a
		reti

; === Bus IO Operations =======================================================
bus_io_table:
		rjmp bus_io_op_console				; IO $00: Console In & Out
		rjmp bus_io_op_constat				; IO $01: Console Status
		rjmp bus_io_op_serial				; IO $02: Serial In & Out
		rjmp bus_io_op_serstat				; IO $03: Serial Status
		rjmp bus_io_op_reserved				; IO $04: (reserved)
		rjmp bus_io_op_reserved				; IO $05: (reserved)
		rjmp bus_io_op_reserved				; IO $06: (reserved)
		rjmp bus_io_op_reserved				; IO $07: (reserved)
		rjmp bus_io_op_disk					; IO $08: Disk I/O (Read/Write)
		rjmp bus_io_op_seldsk				; IO $09: Disk Number
		rjmp bus_io_op_dsktrkhi				; IO $0A: Disk Track (HI Byte)
		rjmp bus_io_op_dsktrklo				; IO $0B: Disk Track (LO Byte)
		rjmp bus_io_op_dsksechi				; IO $0C: Disk Sector (HI Byte)
		rjmp bus_io_op_dskseclo				; IO $0D: Disk Sector (LO Byte)
		rjmp bus_io_op_diskerr				; IO $0E: Disk Status (Read-Only)
		rjmp bus_io_op_buffer				; IO $0F: Buffer In & Out
bus_io_op_reserved:
		ret

; ### IO $00: Console In & Out ################################################
bus_io_op_console:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_console_in
bus_io_op_console_out:						; => Write to Console
		lds a,BUS_IO_DATA
		mcall con_put
		rjmp bus_io_op_console_exit
bus_io_op_console_in:						; => Read from Console
		mcall con_get
		sts BUS_IO_DATA,a

bus_io_op_console_exit:
		pop a
		ret

; ### IO $01: Console Status ###################################################
bus_io_op_constat:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		brne bus_io_op_constat_exit
bus_io_op_constat_get:						; => Returns Console Status
		mcall con_stat
		sts BUS_IO_DATA,a

bus_io_op_constat_exit:
		pop a
		ret

; ### IO $02: Serial In & Out #################################################
bus_io_op_serial:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_serial_in
bus_io_op_serial_out:						; => Write to Serial
		lds a,BUS_IO_DATA
		mcall ser_put
		rjmp bus_io_op_serial_exit
bus_io_op_serial_in:						; => Read from Serial
		mcall ser_get
		sts BUS_IO_DATA,a

bus_io_op_serial_exit:
		pop a
		ret

; ### IO $03: Serial Status ####################################################
bus_io_op_serstat:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		brne bus_io_op_serstat_exit
bus_io_op_serstat_get:						; => Returns Serial Status
		mcall ser_stat
		sts BUS_IO_DATA,a

bus_io_op_serstat_exit:
		pop a
		ret

; ### IO $08: Disk I/O (Read/Write Sector in BUS_BUF) ##########################
bus_io_op_disk:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_disk_read
bus_io_op_disk_write:						; => Write to Disk
		mcall disk_write
		rjmp bus_io_op_disk_exit
bus_io_op_disk_read:						; => Read from Dsk
		mcall disk_read

bus_io_op_disk_exit:
		pop a
		ret

; ### IO $09: Disk Number #####################################################
bus_io_op_seldsk:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_seldsk_get
bus_io_op_seldsk_set:						; => Selects Disk
		lds a,BUS_IO_DATA
		mcall disk_num_set
		rjmp bus_io_op_seldsk_exit
bus_io_op_seldsk_get:						; => Returns Disk Selected
		mcall disk_num_get
		sts BUS_IO_DATA,a

bus_io_op_seldsk_exit:
		pop a
		ret

; ### IO $0A: Disk Track (HI Byte) ############################################
bus_io_op_dsktrkhi:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_dsktrkhi_get
bus_io_op_dsktrkhi_set:						; => Sets Track (HI Byte)
		lds a,BUS_IO_DATA
		mcall disk_track_set_hi
		rjmp bus_io_op_dsktrkhi_exit
bus_io_op_dsktrkhi_get:						; => Returns Track (HI Byte)
		mcall disk_track_get_hi
		sts BUS_IO_DATA,a

bus_io_op_dsktrkhi_exit:
		pop a
		ret
		
; ### IO $0B: Disk Track (LO Byte) ############################################
bus_io_op_dsktrklo:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_dsktrklo_get
bus_io_op_dsktrklo_set:						; => Sets Track (LO Byte)
		lds a,BUS_IO_DATA
		mcall disk_track_set_lo
		rjmp bus_io_op_dsktrklo_exit
bus_io_op_dsktrklo_get:						; => Returns Track (LO Byte)
		mcall disk_track_get_lo
		sts BUS_IO_DATA,a

bus_io_op_dsktrklo_exit:
		pop a
		ret

; ### IO $0C: Disk Sector (HI Byte) ###########################################
bus_io_op_dsksechi:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_dsksechi_get
bus_io_op_dsksechi_set:						; => Sets Sector (HI Byte)
		lds a,BUS_IO_DATA
		mcall disk_sector_set_hi
		rjmp bus_io_op_dsksechi_exit
bus_io_op_dsksechi_get:						; => Returns Sector (HI Byte)
		mcall disk_sector_get_hi
		sts BUS_IO_DATA,a

bus_io_op_dsksechi_exit:
		pop a
		ret
		
; ### IO $0D: Disk Sector (LO Byte) ###########################################
bus_io_op_dskseclo:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_dskseclo_get
bus_io_op_dskseclo_set:						; => Sets Sector (LO Byte)
		lds a,BUS_IO_DATA
		mcall disk_sector_set_lo
		rjmp bus_io_op_dskseclo_exit
bus_io_op_dskseclo_get:						; => Returns Sector (LO Byte)
		mcall disk_sector_get_lo
		sts BUS_IO_DATA,a

bus_io_op_dskseclo_exit:
		pop a
		ret

; ### IO $0E: Last Disk Status (Read-Only) ####################################
bus_io_op_diskerr:
		push a

		lds a,BUS_IO_DIR					; Check Direction
		tst a
		brne bus_io_op_diskerr_exit
bus_io_op_diskerr_get:						; => Returns Last Disk Status
		mcall disk_get_status
		sts BUS_IO_DATA,a

bus_io_op_diskerr_exit:
		pop a
		ret

; ### IO $0F: Buffer In & Out #################################################
bus_io_op_buffer:
		push a
		push zl
		push zh

		lds a,BUS_IO_CMD					; Calculate Pointer in IO Buffer	
		cpi a,BUS_BUF_SIZE
		brsh bus_io_op_buffer_exit
		ldi zl,LOW(BUS_BUF)
		ldi zh,HIGH(BUS_BUF)
		add zl,a
		adc zh,zero
							
		lds a,BUS_IO_DIR					; Check Direction
		tst a
		breq bus_io_op_buffer_in
bus_io_op_buffere_out:						; => Write to Buffer
		lds a,BUS_IO_DATA
		st z,a
		rjmp bus_io_op_buffer_exit
bus_io_op_buffer_in:						; => Read from Buffer
		ld a,z
		sts BUS_IO_DATA,a

bus_io_op_buffer_exit:
		pop zh
		pop zl
		pop a
		ret

; *****************************************************************************
; *** Disk                                                                  ***
; *****************************************************************************

; ### Inits Disk
disk_init:
		push a

		clr a								; Reset internal Vars
		sts DISK_NUM,a 
		sts DISK_TRACK_HI,a
		sts DISK_TRACK_LO,a
		sts DISK_SECTOR_HI,a
		sts DISK_SECTOR_LO,a
		sts DISK_STATUS,a

		pop a
		ret

; ### Gets Last Disk Status
; <a: The Status (see IDE_BIT_ERROR_xxx)
disk_get_status:
		lds a,DISK_STATUS
		ret

; ### Sets Disk Number
; (DISK_STATUS will be updated) 
; >a: The Disk (0-15)
disk_num_set:
		push a

		sts DISK_STATUS,zero				; Reset Disk Status

		cpi a,DISK_NUM_MAX					; Check Range of Disk Number
		brsh disk_num_set_error
		sts DISK_NUM,a						; Save new Disk Number
		rjmp disk_num_set_exit

disk_num_set_error:							; => Error!
		ldi a,IDE_BIT_ERROR_ABRT
		sts DISK_STATUS,a 

disk_num_set_exit:
		pop a
		ret

; ### Gets Disk Number
; <a: The Disk (0-15)
disk_num_get:								
		lds a,DISK_NUM
		ret

; ### Sets Track (No Error Check!)
; >a: The Track (HI/LO)
disk_track_set_hi:
		sts DISK_TRACK_HI,a
		ret
disk_track_set_lo:
		sts DISK_TRACK_LO,a
		ret

; ### Gets Track
; <a: The Track (HI/LO)
disk_track_get_hi:
		lds a,DISK_TRACK_HI
		ret
disk_track_get_lo:
		lds a,DISK_TRACK_LO
		ret

; ### Sets Sector (No Error Check!)
; >a: The Sector (HI/LO)
disk_sector_set_hi:
		sts DISK_SECTOR_HI,a
		ret
disk_sector_set_lo:
		sts DISK_SECTOR_LO,a
		ret

; ### Gets Sector
; <a: The Sector (HI/LO)
disk_sector_get_hi:
		lds a,DISK_SECTOR_HI
		ret
disk_sector_get_lo:
		lds a,DISK_SECTOR_LO
		ret

; ### Write to Disk
; (DISK_STATUS will be updated) 
disk_write:
		push temp
		push a
		push b
		push c
		push d
		push pdat
		push ol
		push oh
		push zl
		push zh

		sts DISK_STATUS,zero				; Reset Disk Status

		mcall disk_calc_lba					; Calculates LBA

		ldi pdat,0							; ### Read current LBA Sector from IDE Master
		mcall ide_read
		tst pdat
		brne disk_write_error

disk_write_update_buf:						; ### Update IDE Buffer from part of CP/M Sector Buffer
		mov temp,a						
		ldi pdat,BUS_BUF_SIZE				; (pdat = Counter for 128 Byte of a CP/M Sector)
		ldi zl,LOW(BUS_BUF)
		ldi zh,HIGH(BUS_BUF)

disk_write_update_buf_loop:
		ld a,z+
		mcall ide_buf_set
		adiw ol:oh,1
		dec pdat
		brne disk_write_update_buf_loop
		mov a,temp

		ldi pdat,0							; ### Writes current LBA Sector to IDE Master
		mcall ide_write
		tst pdat
		breq disk_write_exit				
disk_write_error:							; => Error!
		sts DISK_STATUS,pdat

disk_write_exit:		
		pop zh
		pop zl
		pop oh
		pop ol
		pop pdat
		pop d
		pop c
		pop b
		pop a
		pop temp
		ret

; ### Reads from Disk
; (DISK_STATUS will be updated) 
disk_read:
		push a
		push b
		push c
		push d
		push pdat
		push ol
		push oh
		push zl
		push zh

		sts DISK_STATUS,zero				; Reset Disk Status

		mcall disk_calc_lba					; Calculates LBA
		
		ldi pdat,0							; ### Read current LBA Sector from IDE Master
		mcall ide_read
		tst pdat
		brne disk_read_error

disk_read_update_buf:						; ### Update CP/M Sector Buffer from part of IDE Buffer)
		ldi pdat,BUS_BUF_SIZE				; (pdat = Counter for 128 Byte of a CP/M Sector)
		ldi zl,LOW(BUS_BUF)
		ldi zh,HIGH(BUS_BUF)
disk_read_update_buf_loop:
		mcall ide_buf_get
		st z+,a
		adiw ol:oh,1
		dec pdat
		brne disk_read_update_buf_loop

disk_read_error:							; => Error!
		sts DISK_STATUS,pdat

disk_read_exit:		
		pop zh
		pop zl
		pop oh
		pop ol
		pop pdat
		pop d
		pop c
		pop b
		pop a
		ret

; ### Calculates LBA
; <a: LBA Bits 0-7
; <b: LBA Bits 8-15
; <c: LBA Bits 16-23
; <d: LBA Bits 24-27 
; <o: Index 0-3 (Offset = 128 * Index inside 512 Byte Sector of LBA)
disk_calc_lba:
		ldil DISK_NUM_LBA					; A = (DISK_NUM * DISK_NUM_LBA)
		mcall math_load_a
		ldil 0
		lds a,DISK_NUM
		mcall math_load_o
		mcall math_umul
		ldi a,0								; Save Accu: R[0] = A
		mcall math_var_save_a
		
		ldil DISK_NUM_TRKSECS				; A = (DISK_NUM_TRKSECS * DISK_TRACK)
		mcall math_load_a
		ldil 0						
		lds a,DISK_TRACK_LO
		lds b,DISK_TRACK_HI
		mcall math_load_o
		mcall math_umul

		ldil 0								; A = A + DISK_SECTOR
		lds a,DISK_SECTOR_LO				
		lds b,DISK_SECTOR_HI
		mcall math_load_o
		mcall math_add

		mov oh,a0							; o = (Bits 0-1 of A) * 128 = Offset in 512-Byte LBA Sector
		andi oh,0x03
		clr ol
		lsr oh
		ror ol

		ldi a,2								; A = (A >> 2)
		mcall math_shr

		ldi a,0								; A = A + R[0]
		mcall math_var_load_o
		mcall math_add

		mov a,a0							; Return LBA
		mov b,a1
		mov c,a2
		mov d,a3
		ret

; *****************************************************************************
; *** Console                                                              ***
; *****************************************************************************

; ### Inits Console
con_init:
		push a
		push pdat

		in pdat,PORTC						; Prepare Pins for Console SPI
		sbr pdat,PINC_CCS|PINC_CBUSY|PINC_KCS|PINC_KONOFF
		out PORTC,pdat
		in a,DDRC
		cbr a,PINC_CBUSY|PINC_KONOFF
		sbr a,PINC_CCS|PINC_KCS
		out DDRC,a

		sts CON_KEY,zero					; Reset Key Code
		sts CON_KEY_MAP,zero				; Reset Index for Key Code Mapping

		pop pdat 
		pop a
		ret

; ### Gets Status
; <a: Returns $FF, if Character is ready, otherwise $00, if Character is not available.
con_stat:
		push pdat
		push zl
		push zh

		in pdat,PORTC						; Select Console Keyboard SPI (CSN=0)
		cbr pdat,PINC_KCS
		out PORTC,pdat
		
		clr a								; Send Command for Getting Current Key Code
		out SPDR,a							
con_stat_loop:
		in a,SPSR
		sbrs a,7
		rjmp con_stat_loop
		in a,SPDR

		sbr pdat,PINC_KCS					; Deselect Console LCD SPI (CSN=1)
		out PORTC,pdat

		sts CON_KEY,a						; Store Key Code
		tst a								; Is Key Pressed?
		breq con_stat_no_key
con_stat_chk_map1:							; Should switch Map Index?
		cpi a,0x9A							; Is Key "Setup"?
		brne con_stat_chk_map2
		ldi a,0								; => Yes, Set Index = 0
		sts CON_KEY_MAP,a
		rjmp con_stat_no_key
con_stat_chk_map2:
		cpi a,0x9B							; Is Key "Font"?
		brne con_stat_map
		ldi a,1								; => Yes, Set Index = 0
		sts CON_KEY_MAP,a
		rjmp con_stat_no_key
con_stat_map:								; Map Key Code
		lds a,CON_KEY_MAP
		ldi zl,LOW(2*CON_KEY_MAP_SETUP)
		ldi zh,HIGH(2*CON_KEY_MAP_SETUP)
		cpi a,1
		brne con_stat_map_do
		ldi zl,LOW(2*CON_KEY_MAP_FONT)
		ldi zh,HIGH(2*CON_KEY_MAP_FONT)
con_stat_map_do:	
		lds a,CON_KEY		
		add zl,a
		adc zh,zero
		lpm a,z
		tst a
		breq con_stat_done
		sts CON_KEY,a
con_stat_done:
		ser a								; => Yes, Return $FF
		rjmp con_stat_exit
con_stat_no_key:							; => No, Return $00
		clr a
		sts CON_KEY,zero
con_stat_exit:
		pop zh
		pop zl
		pop pdat
		ret

; ### Reads Char
; <a: The Char (0=No Char)
con_get:
		lds a,CON_KEY
		tst a
		brne con_get_done
		mcall con_stat
		tst a
		breq con_get_no_key
		lds a,CON_KEY
		cpi a,0x80
		brsh con_get_no_key
		rjmp con_get_done
con_get_no_key:								
		clr a
con_get_done:
		sts CON_KEY,zero

		ret

; ### Writes Char
; >a: The Char
con_put: 
		push a
		push pdat

con_put_busy:								; Wait until BUSY is Released
		in pdat,PINC
		andi pdat,PINC_CBUSY
		tst pdat
		brne con_put_busy

		in pdat,PORTC						; Select Console LCD SPI (CSN=0)
		cbr pdat,PINC_CCS
		out PORTC,pdat
		
		out SPDR,a
con_put_loop:
		in a,SPSR
		sbrs a,7
		rjmp con_put_loop				

		sbr pdat,PINC_CCS					; Deselect Console LCD SPI (CSN=1)
		out PORTC,pdat

		pop pdat
		pop a
		ret

; ### Writes New Line (CR & LF)
con_nl:
		push a
		
		ldi a,CHR_LF
		mcall con_put

		pop a
		ret

; ### Writes String
con_str:
		push a
		push b
		push c

		mcall str_get_len
		mov c,a
		tst c
		breq con_str_exit
		clr b
con_str_loop:
		mcall str_get
		cpse a,zero
		mcall con_put
		inc b
		cp b,c
		brlo con_str_loop

con_str_exit:
		pop c
		pop b
		pop a
		ret

; ### Should Reset?
; <a: Returns Flag
con_is_reset:
		push b

		in b,PINC
		andi b,PINC_KONOFF
		clr a
		cpse b,zero
		ser a

		pop b
		ret

; *****************************************************************************
; *** Serial                                                               ***
; *****************************************************************************

; ### Inits Serial Port
ser_init:
		push a

		; TODO

		pop a
		ret

; ### Gets Status
; <a: Returns $FF, if Character is ready, otherwise $00, if Character is not available.
ser_stat:
		push b

		clr a								; Check RXC-Flag
		inreg b,UCSRA
		sbrs b,RXC
		rjmp ser_stat_exit
		ser a

ser_stat_exit:
		pop b
		ret

; ### Reads Char
; <a: The Char (0=No Char)
ser_get:
		rjmp uart_get

; ### Writes Char
; >a: The Char
ser_put: 
		rjmp uart_send

; ### Writes New Line (CR & LF)
ser_nl:
		rjmp uart_send_nl

; ### Writes String
ser_str:
		rjmp uart_send_str

; *****************************************************************************
; *** Debug                                                                ***
; *****************************************************************************

; ### Debugs IDE Buffer to UART
#ifndef _FN__DEBUG_IDE
#define _FN__DEBUG_IDE
_debug_ide:
		push a
		push ol
		push oh
		push zl
		push zh

		clr ol								; Debug IDE Buffer
		clr oh
		ldi zl,LOW(IDE_BUFFER)
		ldi zh,HIGH(IDE_BUFFER)
_debug_ide_loop:
		mov a,ol
		andi a,0x0F
		tst a
		brne _debug_ide_print
		mcall uart_send_nl
		mcall str_clear
		mov a,oh
		mcall str_add_hex
		mov a,ol
		mcall str_add_hex
		ldi a,':'
		mcall str_add
		mcall uart_send_str
_debug_ide_print:
		mcall str_clear
		ldi a,' '
		mcall str_add
		ld a,z+
		mcall str_add_hex
		mcall uart_send_str
		adiw ol:oh,1
		cpi oh,2
		brne _debug_ide_loop
		mcall uart_send_nl

		pop zh
		pop zl
		pop oh
		pop ol
		pop a
		ret
#endif

; ### Debugs BUS Buffer to UART
#ifndef _FN__DEBUG_BUS_BUF
#define _FN__DEBUG_BUS_BUF
_debug_bus_buf:
		push a
		push b
		push zl
		push zh

		clr b								; Debug BUS Buffer
		ldi zl,LOW(BUS_BUF)
		ldi zh,HIGH(BUS_BUF)
_debug_bus_buf_loop:
		mov a,b
		andi a,0x0F
		tst a
		brne _debug_bus_buf_print
		mcall uart_send_nl
		mcall str_clear
		mov a,b
		mcall str_add_hex
		ldi a,':'
		mcall str_add
		mcall uart_send_str
_debug_bus_buf_print:
		mcall str_clear
		ldi a,' '
		mcall str_add
		ld a,z+
		mcall str_add_hex
		mcall uart_send_str
		inc b
		cpi b,BUS_BUF_SIZE
		brne _debug_bus_buf_loop
		mcall uart_send_nl

		pop zh
		pop zl
		pop b
		pop a
		ret
#endif
