#ifndef __MCP23S17_INC
#define __MCP23S17_INC

.include "util.inc"

; #############################################################################
; ###                                                                       ###
; ###                     MCP23S17 Port Expander (via SPI)                  ###  
; ###                                                                       ###
; #############################################################################

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************

.equ MCP23S17_DEV = 0x40					; MCP23S17 DEVICE
.equ MCP23S17_DEV_WRITE = 0					; MCP23S17: WRITE flag
.equ MCP23S17_DEV_READ = 1					; MCP23S17: READ flag

.equ MCP23S17_REG_IODIRA = 0x00				; MCP23S17 [00] IODIRA   = 0xFF (0=Output / 1=Input)
.equ MCP23S17_REG_IODIRB = 0x01				; MCP23S17 [01] IODIRB   = 0xFF (0=Output / 1=Input)
.equ MCP23S17_REG_IPOLA = 0x02				; MCP23S17 [02] IPOLA    = 0x00 (0=Same logic state of the input pin / 1=Opposite logic state of the input pin)
.equ MCP23S17_REG_IPOLB = 0x03				; MCP23S17 [03] IPOLB    = 0x00 (0=Same logic state of the input pin / 1=Opposite logic state of the input pin) 
.equ MCP23S17_REG_GPINTENA = 0x04			; MCP23S17 [04] GPINTENA = 0x00 (0=Disable GPIO input pin for interrupt-on-change event / 1=Enable GPIO input pin for interrupt-on-change event)
.equ MCP23S17_REG_GPINTENB = 0x05			; MCP23S17 [05] GPINTENB = 0x00 (0=Disable GPIO input pin for interrupt-on-change event / 1=Enable GPIO input pin for interrupt-on-change event)
.equ MCP23S17_REG_DEFVALA = 0x06			; MCP23S17 [06] DEFVALA  = 0x00 (Compare value for pins configured for interrupt-on-change, see INTCON and GPINTEN)
.equ MCP23S17_REG_DEFVALB = 0x07			; MCP23S17 [07] DEFVALB  = 0x00 (Compare value for pins configured for interrupt-on-change, see INTCON and GPINTEN)
.equ MCP23S17_REG_INTCONA = 0x08			; MCP23S17 [08] INTCONA  = 0x00 (0=Pin value is compared against the previous pin value / 1=Controls how the associated pin value is compared for interrupt-on-change)
.equ MCP23S17_REG_INTCONB = 0x09			; MCP23S17 [09] INTCONB  = 0x00 (0=Pin value is compared against the previous pin value / 1=Controls how the associated pin value is compared for interrupt-on-change)
.equ MCP23S17_REG_IOCON = 0x0A				; MCP23S17 [0A] IOCON    = 0x00 (see MCP23S17_IOCON_xxx definitions)
.equ MCP23S17_REG_GPPUA = 0x0C				; MCP23S17 [0C] GPPUA    = 0x00 (0=Pull-up disabled / 1=Pull-up enabled)
.equ MCP23S17_REG_GPPUB = 0x0D				; MCP23S17 [0D] GPPUB    = 0x00 (0=Pull-up disabled / 1=Pull-up enabled)
.equ MCP23S17_REG_INTFA	= 0x0E				; MCP23S17 [0E] INTFA    = 0x00 (INTERRUPT FLAG REGISTER: 0=Interrupt not pending / 1=Pin caused interrupt), read-only
.equ MCP23S17_REG_INTFB = 0x0F				; MCP23S17 [0F] INTFB    = 0x00 (INTERRUPT FLAG REGISTER: 0=Interrupt not pending / 1=Pin caused interrupt), read-only
.equ MCP23S17_REG_INTCAPA = 0x10			; MCP23S17 [10] INTCAPA  = 0x00 (INTERRUPT CAPTURE REGISTER: 0=Logic-low / 1=Logic-high), read-only
.equ MCP23S17_REG_INTCAPB = 0x11			; MCP23S17 [11] INTCAPB  = 0x00 (INTERRUPT CAPTURE REGISTER: 0=Logic-low / 1=Logic-high), read-only
.equ MCP23S17_REG_GPIOA = 0x12				; MCP23S17 [12] GPIOA    = 0x00 (PORT REGISTER: 0=Logic-low / 1=Logic-high)
.equ MCP23S17_REG_GPIOB = 0x13				; MCP23S17 [13] GPIOB    = 0x00 (PORT REGISTER: 0=Logic-low / 1=Logic-high)
.equ MCP23S17_REG_OLATA = 0x14				; MCP23S17 [14] OLATA    = 0x00 (OUTPUT LATCH REGISTER: 0=Logic-low / 1=Logic-high)
.equ MCP23S17_REG_OLATB = 0x15				; MCP23S17 [15] OLATB    = 0x00 (OUTPUT LATCH REGISTER: 0=Logic-low / 1=Logic-high)

.equ MCP23S17_IOCON_BANK = (1<<7)			; MCP23S17 IOCON REGISTER Bit 7 (BANK): 0=All registers in same bank (addresses are sequential) / 1=Registers are separated into different banks by ports
.equ MCP23S17_IOCON_MIRROR = (1<<6)			; MCP23S17 IOCON REGISTER Bit 6 (MIRROR): 0=INT pins are not connected. INTA is associated with PortA and INTB is associated with PortB / 1=INT pins are internally connected
.equ MCP23S17_IOCON_SEQOP = (1<<5)			; MCP23S17 IOCON REGISTER Bit 5 (SEQOP): 0=Sequential operation enabled, address pointer increments / 1=Sequential operation disabled, address pointer does not increment
.equ MCP23S17_IOCON_DISSLW = (1<<4)			; MCP23S17 IOCON REGISTER Bit 4 (DISSLW): 0=Slew rate enabled / 1=Slew rate disabled
.equ MCP23S17_IOCON_HAEN = (1<<3)			; MCP23S17 IOCON REGISTER Bit 3 (HAEN): 0=Disables the MCP23S17 address pins (Default address is 0) / 1=Enables the MCP23S17 address pins
.equ MCP23S17_IOCON_ODR = (1<<2)			; MCP23S17 IOCON REGISTER Bit 2 (ODR): 0=Active driver output (INTPOL bit sets the polarity) / 1=Open-drain output (overrides the INTPOL bit)
.equ MCP23S17_IOCON_INTPOL = (1<<1)			; MCP23S17 IOCON REGISTER Bit 1 (INTPOL): 0=Active-low / 1=Active-high

.equ MCP23S17_ADDR0	= 0x00					; MCP23S17 DEVICE ADDR #0
.equ MCP23S17_ADDR1 = 0x20					; MCP23S17 DEVICE ADDR #1
.equ MCP23S17_ADDR2 = 0x40					; MCP23S17 DEVICE ADDR #2
.equ MCP23S17_ADDR3 = 0x60					; MCP23S17 DEVICE ADDR #3
.equ MCP23S17_ADDR4 = 0x80					; MCP23S17 DEVICE ADDR #4
.equ MCP23S17_ADDR5 = 0xA0					; MCP23S17 DEVICE ADDR #5
.equ MCP23S17_ADDR6 = 0xC0					; MCP23S17 DEVICE ADDR #6
.equ MCP23S17_ADDR7 = 0xE0					; MCP23S17 DEVICE ADDR #7

; ### PINS
; (using default hardware SPI)
#ifndef MCP23S17_PINS
#define MCP23S17_PINS
.equ MCP23S17_PIN_CSN = (1<<2)				; PB2: CSN
.equ MCP23S17_PIN_MOSI = (1<<3)				; PB3: MOSI 
.equ MCP23S17_PIN_MISO = (1<<4)				; PB4: MISO
.equ MCP23S17_PIN_SCK = (1<<5)				; PB5: SCK 
.equ MCP23S17_PORTDAT = PORTB
.equ MCP23S17_PORTDIR = DDRB
#endif

; ### HW SPI SPEED
#ifndef MCP23S17_SPI_SPEED
#define MCP23S17_SPI_SPEED
.equ MCP23S17_SPR1 = (0<<SPR0)|(0<<SPR1)
.equ MCP23S17_SPR2 = (1<<SPI2X)
#endif

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

.cseg
; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

; *****************************************************************************
; *** Public functions                                                      ***
; *****************************************************************************

; ### Inits MCP23S17
#ifndef MCP23S17_FN_INIT
#define MCP23S17_FN_INIT
mcp23s17_init:
		push a
		push b
		push pdat

		lds a,UTIL_INIT1					; Check, if initialization has been already done
		sbrc a,INIT1_MCP23S17				
		rjmp mcp23s17_init_done

		mcall util_init						; Init depending modules
		
		in a,MCP23S17_PORTDIR				; Prepare pins
		cbr a,MCP23S17_PIN_CSN|MCP23S17_PIN_MOSI|MCP23S17_PIN_MISO|MCP23S17_PIN_SCK
		sbr a,MCP23S17_PIN_CSN|MCP23S17_PIN_MOSI|MCP23S17_PIN_SCK
		out MCP23S17_PORTDIR,a
		in pdat,MCP23S17_PORTDAT
		cbr pdat,MCP23S17_PIN_CSN|MCP23S17_PIN_MOSI|MCP23S17_PIN_SCK
		sbr pdat,MCP23S17_PIN_CSN|MCP23S17_PIN_MISO
		out MCP23S17_PORTDAT,pdat						

		clr a								; Configure SPI
		sbr a,(0<<SPIE)						; SPIE (SPI interrupt enable)
		sbr a,(1<<SPE)						; SPE (SPI enable)
		sbr a,(0<<DORD)						; DORD (Data order)				-> 0: MSB of the data word is transmitted first / 1: LSB of the data word is transmitted first.
		sbr a,(1<<MSTR)						; MSTR (Master/slave select)	-> 0: Slave / 1: Master 
		sbr a,(0<<CPOL)						; CPOL (Clock polarity)			-> 0: Leading edge = Rising + Trailing edge = Falling (SCK low when idle) / 1: Leading edge = Falling + Trailing edge = Rising
		sbr a,(0<<CPHA)						; CPHA (Clock phase)			-> 0: Leading edge = Sample + Trailing edge = Setup / 1: Leading edge = Setup + Trailing edge = Sample
		sbr a,MCP23S17_SPR1					; SPRx (SPI clock rate select)	-> 000: fosc/4, 001: fosc/16, 010: fosc/64, 011: fosc/128, 100: fosc/2, 101: fosc/8, 110: fosc/32, 111: fosc/64
		out SPCR,a

		clr a
		sbr a,MCP23S17_SPR2					; SPI2X: Double SPI speed bit (see SPRx)
		out SPSR,a

mcp23s17_init_haen:							; "Hardware Address Enable" for all devices
		ldi b,MCP23S17_ADDR0|MCP23S17_REG_IOCON		
		rcall mcp23s17_read
		sbr a,MCP23S17_IOCON_HAEN
		rcall mcp23s17_write

mcp23s17_init_done:
		lds a,UTIL_INIT1					; Mark as initialized
		sbr a,(1<<INIT1_MCP23S17)
		sts UTIL_INIT1,a

		pop pdat
		pop b
		pop a
		ret
#endif

; ### Gets I/O direction of GPIO-A (see MCP23S17_REG_IODIRA)
; <a: The direction of the ports (0=output/1=input)
; >b: Select address
#ifndef MCP23S17_FN_GET_IODIR_A
#define MCP23S17_FN_GET_IODIR_A
mcp23s17_get_iodir_a:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_IODIRA
		rcall mcp23s17_read					; Get register IODIRA
			
		pop b
		ret
#endif

; ### Sets I/O direction of GPIO-A (see MCP23S17_REG_IODIRA)
; >a: The direction of the ports (0=output/1=input)
; >b: Select address
#ifndef MCP23S17_FN_SET_IODIR_A
#define MCP23S17_FN_SET_IODIR_A
mcp23s17_set_iodir_a:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_IODIRA
		rcall mcp23s17_write				; Set register IODIRA
			
		pop b
		ret
#endif

; ### Gets I/O direction of GPIO-B (see MCP23S17_REG_IODIRB)
; <a: The direction of the ports (0=output/1=input)
; >b: Select address
#ifndef MCP23S17_FN_GET_IODIR_B
#define MCP23S17_FN_GET_IODIR_B
mcp23s17_get_iodir_b:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_IODIRB
		rcall mcp23s17_read					; Get register IODIRB
			
		pop b
		ret
#endif

; ### Sets I/O direction of GPIO-B (see MCP23S17_REG_IODIRB)
; >a: The direction of the ports (0=output/1=input)
; >b: Select address
#ifndef MCP23S17_FN_SET_IODIR_B
#define MCP23S17_FN_SET_IODIR_B
mcp23s17_set_iodir_b:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_IODIRB
		rcall mcp23s17_write				; Set register IODIRB
			
		pop b
		ret
#endif

; ### Sets pull-up for input of GPIO-A (see MCP23S17_REG_GPPUA)
; <a: The pull-up flags of input pins (0=disabled/1=enabled)
; >b: Select address
#ifndef MCP23S17_FN_SET_GPPU_A
#define MCP23S17_FN_SET_GPPU_A
mcp23s17_set_pullup_a:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_GPPUA
		rcall mcp23s17_write				; Set register GPPUA
			
		pop b
		ret
#endif

; ### Sets pull-up for input of GPIO-B (see MCP23S17_REG_GPPUB)
; <a: The pull-up flags of input pins (0=disabled/1=enabled)
; >b: Select address
#ifndef MCP23S17_FN_SET_GPPU_B
#define MCP23S17_FN_SET_GPPU_B
mcp23s17_set_pullup_b:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_GPPUB
		rcall mcp23s17_write				; Set register GPPUB
			
		pop b
		ret
#endif

; ### Gets port of GPIO-A (see MCP23S17_REG_GPIOA)
; <a: The logical state of the port pins 
; >b: Select address
#ifndef MCP23S17_FN_GET_GPIO_A
#define MCP23S17_FN_GET_GPIO_A
mcp23s17_get_gpio_a:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_GPIOA
		rcall mcp23s17_read					; Get register GPIOA
			
		pop b
		ret
#endif

; ### Sets port of GPIO-A (see MCP23S17_REG_GPIOA)
; >a: The logical state of the port pins 
; >b: Select address
#ifndef MCP23S17_FN_SET_GPIO_A
#define MCP23S17_FN_SET_GPIO_A
mcp23s17_set_gpio_a:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_GPIOA
		rcall mcp23s17_write				; Set register GPIOA
			
		pop b
		ret
#endif

; ### Gets port of GPIO-B (see MCP23S17_REG_GPIOB)
; <a: The logical state of the port pins 
; >b: Select address
#ifndef MCP23S17_FN_GET_GPIO_B
#define MCP23S17_FN_GET_GPIO_B
mcp23s17_get_gpio_b:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_GPIOB
		rcall mcp23s17_read					; Get register GPIOB
			
		pop b
		ret
#endif

; ### Sets port of GPIO-B (see MCP23S17_REG_GPIOB)
; >a: The logical state of the port pins 
; >b: Select address
#ifndef MCP23S17_FN_SET_GPIO_B
#define MCP23S17_FN_SET_GPIO_B
mcp23s17_set_gpio_b:
		push b
		
		andi b,0xE0							; Prepare address
		sbr b,MCP23S17_REG_GPIOB
		rcall mcp23s17_write				; Set register GPIOB
			
		pop b
		ret
#endif

; ### Reads byte from register
; <a: The byte read
; >b: Bit0-4: The register & Bit5-7: Select address
#ifndef MCP23S17_FN_READ
#define MCP23S17_FN_READ
mcp23s17_read:
		push c
		push pdat

		in pdat,MCP23S17_PORTDAT			; Select device (CSN=0)
		cbr pdat,MCP23S17_PIN_CSN
		out MCP23S17_PORTDAT,pdat
		
		mov c,b								; Send device & address via SPI
		swap c
		andi c,0x0E
		ori c,MCP23S17_DEV|MCP23S17_DEV_READ
		out SPDR,c
mcp23s17_read_loop1:
		in c,SPSR
		sbrs c,7
		rjmp mcp23s17_read_loop1				

		mov c,b								; Send register address via SPI
		andi c,0x1F
		out SPDR,c
mcp23s17_read_loop2:
		in c,SPSR
		sbrs c,7
		rjmp mcp23s17_read_loop2			

		out SPDR,zero						; Send dummy byte
mcp23s17_read_loop3:
		in c,SPSR
		sbrs c,7
		rjmp mcp23s17_read_loop3			

		sbr pdat,MCP23S17_PIN_CSN			; Deselect device (CSN=1)
		out MCP23S17_PORTDAT,pdat

		in a,SPDR							; Read value

		pop pdat
		pop c
		ret
#endif

; ### Writes byte to register
; <a: The byte to write
; >b: Bit0-4: The register & Bit5-7: Select address
#ifndef MCP23S17_FN_WRITE
#define MCP23S17_FN_WRITE
mcp23s17_write:
		push c
		push pdat

		in pdat,MCP23S17_PORTDAT			; Select device (CSN=0)
		cbr pdat,MCP23S17_PIN_CSN
		out MCP23S17_PORTDAT,pdat
		
		mov c,b								; Send device & address via SPI
		swap c
		andi c,0x0E
		ori c,MCP23S17_DEV|MCP23S17_DEV_WRITE
		out SPDR,c
mcp23s17_write_loop1:
		in c,SPSR
		sbrs c,7
		rjmp mcp23s17_write_loop1				

		mov c,b								; Send register address via SPI
		andi c,0x1F
		out SPDR,c
mcp23s17_write_loop2:
		in c,SPSR
		sbrs c,7
		rjmp mcp23s17_write_loop2			

		out SPDR,a							; Send value
mcp23s17_write_loop3:
		in c,SPSR
		sbrs c,7
		rjmp mcp23s17_write_loop3			

		sbr pdat,MCP23S17_PIN_CSN			; Deselect device (CSN=1)
		out MCP23S17_PORTDAT,pdat

		pop pdat
		pop c
		ret
#endif

; *****************************************************************************
; *** Private functions                                                     ***
; *****************************************************************************

#endif ; __MCP23S17_INC
