#ifndef __STR_INC
#define __STR_INC

.include "util.inc"

; #############################################################################
; ###                                                                       ###
; ###                            String Utility                             ###  
; ###                                                                       ###
; #############################################################################

; !!! Use define STR_DEC_IGNORE for ignoring some decimal functions !!!

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************

; === Format Identifiers (Placeholders) =======================================

.equ STR_FMTID_CHR = 1						; Argument as ASCII Character
.equ STR_FMTID_HEX = 2						; Argument as Hexadecimal Byte (2 Digits)
.equ STR_FMTID_DEC = 3						; Argument as Decimal Byte (1-3 Digits)
.equ STR_FMTID_DEC2 = 4						; Argument as Decimal Byte (always 2 Digits)
.equ STR_FMTID_DEC3 = 5						; Argument as Decimal Byte (always 3 Digits)
.equ STR_FMTID_DEC_WORD = 6					; Argument as Decimal Word (1-5 Digits), Word have two Bytes in HI-LO Order 
.equ STR_FMTID_DEC_WORD3 = 7				; Argument as Decimal Word (3 Digits) 
.equ STR_FMTID_DEC_WORD4 = 8				; Argument as Decimal Word (4 Digits) 
.equ STR_FMTID_DEC_WORD5 = 9				; Argument as Decimal Word (5 Digits) 
.equ STR_FMTID_BIT = 10						; Argument as Bits (8 Digits)

; === CONSTANTS ===============================================================

.ifndef STR_BUF_SIZE						; Size of buffer
.equ STR_BUF_SIZE = 80
.endif
.ifndef STR_FMT_SIZE						; Size of format arguments
.equ STR_FMT_SIZE = 16		
.endif

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

STR_BUF:	.byte STR_BUF_SIZE				; String Buffer
.if STR_FMT_SIZE>0
STR_FMT:	.byte STR_FMT_SIZE				; Buffer for Format Arguments
.endif
STR_LEN:	.byte 1							; String Length inside Buffer
STR_NUM:	.byte 1							; Number of Format Arguments

.cseg
; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

; *****************************************************************************
; *** Public functions                                                      ***
; *****************************************************************************

; ### Inits String Module
#ifndef STR_FN_INIT
#define STR_FN_INIT
str_init:
		push a
		
		sbrc init0,INIT0_STR				; Check, if initialization has been already done
		rjmp str_init_done
		
		mcall util_init						; Init depending modules

.if STR_FMT_SIZE > 0
		rcall str_fmt_reset					; Resets Buffer for Format Arguments
.endif
		rcall str_clear						; Clear Buffer	

str_init_done:
		mov a,init0							; Mark as initialized
		sbr a,(1<<INIT0_STR)
		mov init0,a

		pop a
		ret
#endif

; ### Clears Buffer
#ifndef STR_FN_CLEAR
#define STR_FN_CLEAR
str_clear:
		push a
		push b
		push zl
		push zh

		clr a								; Reset String Length
		sts STR_LEN,a
		ldi zl,LOW(STR_BUF)					; Clear Buffer
		ldi zh,HIGH(STR_BUF)
		ldi b,STR_BUF_SIZE
str_clear_loop:
		st z+,a
		dec b
		brne str_clear_loop

		pop zh
		pop zl
		pop b
		pop a
		ret
#endif

; ### Pushes String from SRAM to buffer
; a> Number of Chars to copy
; z> Source pointer to copy from
#ifndef STR_FN_PUSH
#define STR_FN_PUSH
str_push: 
		push a
		push b
		push xl
		push xh
		push zl
		push zh
		
		rcall str_clear						; Clear String Buffer	

		tst a								; Check Num Of Chars
		breq str_push_exit
		ldi xl,LOW(STR_BUF)					; Copies String
		ldi xh,HIGH(STR_BUF)
		mov b,a
str_push_loop:
		ld a,z+
		rcall str_add
		dec b
		brne str_push_loop

str_push_exit:
		pop zh
		pop zl
		pop xh
		pop xl
		pop b
		pop a
		ret
#endif

; ### Gets String Length
; <a: The Length
#ifndef STR_FN_GET_LEN
#define STR_FN_GET_LEN
str_get_len:
		lds a,STR_LEN
		ret
#endif

; ### Gets Char from Buffer
; <a: The Char
; >b: The Position 
#ifndef STR_FN_GET
#define STR_FN_GET
str_get:
		push zl
		push zh

		clr a

		cpi b,STR_BUF_SIZE					; Check Position
		brsh str_get_exit
		ldi zl,LOW(STR_BUF)					; Read Char
		ldi zh,HIGH(STR_BUF)
		add zl,b
		adc zh,zero
		ld a,z
		
str_get_exit:
		pop zh
		pop zl
		ret
#endif

; ### Sets Char into Buffer
; >a: The Char
; >b: The Position 
#ifndef STR_FN_SET
#define STR_FN_SET
str_set:
		push a
		push b
		push zl
		push zh

		cpi b,STR_BUF_SIZE					; Check Position
		brsh str_get_exit
		ldi zl,LOW(STR_BUF)					; Write Char
		ldi zh,HIGH(STR_BUF)
		add zl,b
		adc zh,zero
		st z,a
		inc b								; Syncronize String Length
		lds a,STR_LEN						
		cp a,b
		brsh str_set_exit
		sts STR_LEN,b

str_set_exit:
		pop zh
		pop zl
		pop b
		pop a
		ret
#endif

; ### Adds Char to Buffer
; >a: The Char
#ifndef STR_FN_ADD
#define STR_FN_ADD
str_add:
		push b
		push zl
		push zh

		lds b,STR_LEN						; Check Free Slot
		cpi b,STR_BUF_SIZE
		brsh str_add_exit
		ldi zl,LOW(STR_BUF)					; Write Char
		ldi zh,HIGH(STR_BUF)
		add zl,b
		adc zh,zero
		st z,a
		inc b								; Increment String Length
		sts STR_LEN,b

str_add_exit:
		pop zh
		pop zl
		pop b
		ret
#endif


; ### Resets Format Arguments
#ifndef STR_FN_FMT_RESET
#define STR_FN_FMT_RESET
.if STR_FMT_SIZE > 0
str_fmt_reset:
		push a
		push b
		push zl
		push zh

		clr a								; Reset Number of Arguments
		sts STR_NUM,a
		ldi zl,LOW(STR_FMT)					; Clear Buffer
		ldi zh,HIGH(STR_FMT)
		ldi b,STR_FMT_SIZE
str_fmt_reset_loop:
		st z+,a
		dec b
		brne str_fmt_reset_loop

		pop zh
		pop zl
		pop b
		pop a
		ret
.endif
#endif

; ### Gets Data from String Format
; <a: The Data Byte
; >b: The Position 
#ifndef STR_FN_FMT_GET
#define STR_FN_FMT_GET
.if STR_FMT_SIZE > 0
str_fmt_get:
		push zl
		push zh

		clr a

		cpi b,STR_FMT_SIZE					; Check Position
		brsh str_fmt_get_exit
		ldi zl,LOW(STR_FMT)					; Read Data Byte
		ldi zh,HIGH(STR_FMT)
		add zl,b
		adc zh,zero
		ld a,z
		
str_fmt_get_exit:
		pop zh
		pop zl
		ret
.endif
#endif

; ### Adds Format Argument 
; >a: The Argument
#ifndef STR_FN_FMT_ADD
#define STR_FN_FMT_ADD
.if STR_FMT_SIZE > 0
str_fmt_add:
		push b
		push zl
		push zh

		lds b,STR_NUM						; Check Free Slot for Agument
		cpi b,STR_FMT_SIZE
		brsh str_fmt_add_exit
		ldi zl,LOW(STR_FMT)					; Write Argument
		ldi zh,HIGH(STR_FMT)
		add zl,b
		adc zh,zero
		st z,a
		inc b								; Increment Num Of Arguments
		sts STR_NUM,b

str_fmt_add_exit:
		pop zh
		pop zl
		pop b
		ret
.endif
#endif

; ### Adds String (Unformatted) from EEPROM
#ifndef STR_ADD_STR
#define STR_ADD_STR
str_add_str:
		push a
		push zl
		push zh

str_add_str_loop:							; Read current Character from EEPROM
		mcall util_get_eeprom
		tst a
		breq str_add_str_exit
		rcall str_add						; Add Character
		adiw z,1							; Do next Character
		rjmp str_add_str_loop

str_add_str_exit:
		pop zh
		pop zl
		pop a
		ret
#endif

; ### Generates Formatted String from EEPROM
; >z: Pointer inside PROM
#ifndef STR_ADD_FMT
#define STR_ADD_FMT
.if STR_FMT_SIZE > 0
str_add_fmt:
		push a
		push b
		push c
		push d
		push ol
		push oh
		push xl
		push xh
		push yl
		push yh
		push zl
		push zh
		
		mov yl,zl
		mov yh,zh
		ldi xl,LOW(STR_FMT)
		ldi xh,HIGH(STR_FMT)
str_add_fmt_loop:							; Read current Character from EEPROM
		mov zl,yl
		mov zh,yh
		mcall util_get_eeprom
		tst a
		breq str_add_fmt_exit
		cpi a,' '
		brsh str_add_fmt_add_chr
		ldi zl,LOW(str_add_fmt_table)		; Add formatted Data
		ldi zh,HIGH(str_add_fmt_table)
		add zl,a
		adc zh,zero
		icall		
		rjmp str_add_fmt_next
str_add_fmt_add_chr:						; Add Character
		rcall str_add
str_add_fmt_next:
		adiw y,1							; Do next Character in Format String
		rjmp str_add_fmt_loop

str_add_fmt_exit:
		pop zh
		pop zl
		pop yh
		pop yl
		pop xh
		pop xl
		pop oh
		pop ol
		pop d
		pop c
		pop b
		pop a
		ret
str_add_fmt_table:
		rjmp str_add_fmt_id_none
		rjmp str_add_fmt_id_chr				; 01) Format Identifier as ASCII Character
		rjmp str_add_fmt_id_hex				; 02) Format Identifier as Hexadecimal Byte (2 Digits)
#ifndef STR_DEC_IGNORE
		rjmp str_add_fmt_id_dec				; 03) Format Identifier as Decimal Byte (1-3 Digits)
		rjmp str_add_fmt_id_dec2			; 04) Format Identifier as Decimal Byte (always 2 Digits)
		rjmp str_add_fmt_id_dec3			; 05) Format Identifier as Decimal Byte (always 3 Digits)
		rjmp str_add_fmt_id_dec_word		; 06) Format Identifier as Decimal Word (1-5 Digits)
		rjmp str_add_fmt_id_dec_word3		; 07) Format Identifier as Decimal Word (always 3 Digits)
		rjmp str_add_fmt_id_dec_word4		; 08) Format Identifier as Decimal Word (always 4 Digits)
		rjmp str_add_fmt_id_dec_word5		; 09) Format Identifier as Decimal Word (always 5 Digits)
#else
		rjmp str_add_fmt_id_none
		rjmp str_add_fmt_id_none
		rjmp str_add_fmt_id_none
		rjmp str_add_fmt_id_none
		rjmp str_add_fmt_id_none
		rjmp str_add_fmt_id_none
		rjmp str_add_fmt_id_none
#endif
		rjmp str_add_fmt_id_bit				; 10) Format Identifier as Bits (always 8 Digits)
str_add_fmt_id_none:
		ret
str_add_fmt_id_chr:							; 1) Format Identifier as ASCII Character
		ld a,x+
		cpse a,zero
		rcall str_add
		ret
str_add_fmt_id_hex:							; 2) Format Identifier as Hexadecimal Byte (2 Digits)
		ld a,x+
		rcall str_add_hex
		ret
#ifndef STR_DEC_IGNORE
str_add_fmt_id_dec:							; 3) Format Identifier as Decimal Byte (1-3 Digits)
		ld a,x+
		rcall str_add_dec
		ret
str_add_fmt_id_dec2:						; 4) Format Identifier as Decimal Byte (always 2 Digits)
		ld a,x+
		mcall util_byte2bcd
str_add_fmt_id_dec2_add:
		mov a,c
		adi a,'0'
		rcall str_add
		mov a,d
		adi a,'0'
		rcall str_add
		ret
str_add_fmt_id_dec3:						; 5) Format Identifier as Decimal Byte (always 3 Digits)
		ld a,x+
		mcall util_byte2bcd
		mov a,b
		adi a,'0'
		rcall str_add
		rjmp str_add_fmt_id_dec2_add
str_add_fmt_id_dec_word:					; 6) Format Identifier as Decimal Word (1-5 Digits)
		ld oh,x+
		ld ol,x+
		rcall str_add_dec_word
		ret
str_add_fmt_id_dec_word3:					; 7) Format Identifier as Decimal Word (always 3 Digits)
		ld oh,x+
		ld ol,x+
		mcall util_word2bcd
str_add_fmt_id_dec_word3_add:
		lds a,UTIL_TMP_DIGITS+2
		adi a,'0'
		rcall str_add
		lds a,UTIL_TMP_DIGITS+3
		adi a,'0'
		rcall str_add
		lds a,UTIL_TMP_DIGITS+4
		adi a,'0'
		rcall str_add
		ret
str_add_fmt_id_dec_word4:					; 8) Format Identifier as Decimal Word (always 4 Digits)
		ld oh,x+
		ld ol,x+
		mcall util_word2bcd
str_add_fmt_id_dec_word4_add:
		lds a,UTIL_TMP_DIGITS+1
		adi a,'0'
		rcall str_add
		rjmp str_add_fmt_id_dec_word3_add
str_add_fmt_id_dec_word5:					; 9) Format Identifier as Decimal Word (always 5 Digits)
		ld oh,x+
		ld ol,x+
		mcall util_word2bcd
		lds a,UTIL_TMP_DIGITS
		adi a,'0'
		rcall str_add
		rjmp str_add_fmt_id_dec_word4_add
#endif
str_add_fmt_id_bit:
		ld a,x+
		rcall str_add_bit
		ret
.endif
#endif

; ### Adds Byte as Hexadecimal to Buffer
; >a: The Byte
#ifndef STR_ADD_FN_HEX
#define STR_ADD_FN_HEX
str_add_hex:
		push a
		push b
		push c

		mcall util_byte2hex
		mov a,b
		rcall str_add
		mov a,c
		rcall str_add

		pop c
		pop b
		pop a
		ret
#endif

; ### Adds Byte as Decimal (1-3 digits) to Buffer
; >a: The Byte
#ifndef STR_FN_ADD_DEC
#define STR_FN_ADD_DEC
#ifndef STR_DEC_IGNORE
str_add_dec:
		push ol
		push oh
		
		mov ol,a							
		clr oh
		rcall str_add_dec_word

		pop oh
		pop ol
		ret
#endif
#endif

; ### Adds 16-bit Word as Decimal (1-5 digits) to Buffer
; >o: The 16bit-Word to send
#ifndef STR_FN_ADD_DEC_WORD
#define STR_FN_ADD_DEC_WORD
#ifndef STR_DEC_IGNORE
str_add_dec_word:
		push a
		push b
		push xl
		push xh

		mcall util_word2dec					; Convert to ASCII

		ldi xl,LOW(UTIL_TMP_DIGITS)			; Add Digits, except spaces ' '
		ldi xh,HIGH(UTIL_TMP_DIGITS)
		ldi b,5
str_add_dec_word_loop:
		ld a,x+
		cpi a,' '
		breq str_add_dec_word_next
		rcall str_add
str_add_dec_word_next:
		dec b
		brne str_add_dec_word_loop

		pop xh
		pop xl
		pop b
		pop a
		ret
#endif
#endif

; ### Adds Byte as Binary (8 Digits for each Bit, MSB First) to Buffer
; >a: The Byte
#ifndef STR_FN_ADD_BIT
#define STR_FN_ADD_BIT
str_add_bit:
		push a 
		push b
		push c
		
		mov b,a
		ldi c,8
str_add_bit_loop:
		ldi a,'0'
		lsl b
		brcc str_add_bit_cont
		ldi a,'1'
str_add_bit_cont:
		rcall str_add
		dec c
		brne str_add_bit_loop

		pop c
		pop b
		pop a
		ret
#endif

; ### Reads Hexadecimal Byte from String
; <a: The resulting value as output
; >b: The Start Position
#ifndef STR_FN_READ_HEX
#define STR_FN_READ_HEX
str_read_hex:
		push b
		push c
		push d
		push e
		push zl
		push zh

		clr d								; Temporary Digit #1
		clr e								; Temporary Digit #2
				
		lds c,STR_LEN						; Is start position inside range?
		cp b,c								
		brsh str_read_hex_convert

str_read_hex_start:
		rcall str_get						; Digit #1?
		mov d,a
		inc b
		cp b,c
		brsh str_read_hex_convert
		rcall str_get						; Digit #2?
		mov e,a
		
str_read_hex_convert:
		mov b,d
		mov c,e
		mcall util_hex2byte		

str_read_hex_exit:
		pop zh
		pop zl
		pop e
		pop d
		pop c
		pop b
		ret
#endif 

; ### Reads Decimal Byte from String
; <<carry: If carry-flag is set, there is an invalid character
; <a: The resulting value as output
; >b: The Start Position
#ifndef STR_FN_READ_DEC
#define STR_FN_READ_DEC
#ifndef STR_DEC_IGNORE
str_read_dec:
		push b
		push c
		push d
		push e
		push f
		push zl
		push zh

		clr d								; Temporary Digit #1
		clr e								; Temporary Digit #2
		clr f								; Temporary Digit #3
		
		lds c,STR_LEN						; Is start position inside range?
		cp b,c								
		brsh str_read_dec_failed
str_read_dec_start:
		rcall str_get						; Digit #1 (100)?
		cpi a,'0'
		brlo str_read_dec_failed
		cpi a,'0'+10
		brsh str_read_dec_failed
		mov d,a	
		inc b
		cp b,c
		brsh str_read_dec_done
		rcall str_get						; Digit #2 (10)?
		cpi a,'0'							
		brlo str_read_dec_done
		cpi a,'0'+10
		brsh str_read_dec_done
		mov e,a	
		inc b
		cp b,c
		brsh str_read_dec_done
		rcall str_get						; Digit #3 (1)?
		cpi a,'0'							
		brlo str_read_dec_done
		cpi a,'0'+10
		brsh str_read_dec_done
		mov f,a	

str_read_dec_done:
		mov b,d
		mov c,e
		mov d,f

str_read_dec_arrange:
		tst d								; Arrange Digits (1-3 Digits are used)
		brne str_read_dec_convert
		mov d,c
		mov c,b
		ldi b,'0'
		rjmp str_read_dec_arrange
str_read_dec_convert:
		mcall util_dec2byte		
		rjmp str_read_dec_exit
str_read_dec_failed:
		sec
str_read_dec_exit:
		pop zh
		pop zl
		pop f
		pop e
		pop d
		pop c
		pop b
		ret
#endif
#endif

; ### Finds Character inside String
; <>a: Start position as input and position found on output (or $FF, if not found)
; >b: The character to find
#ifndef STR_FN_FIND_CHR
#define STR_FN_FIND_CHR
str_find_chr:
		push temp
		push b
		push c
		push zl
		push zh
		
		lds c,STR_LEN						; Is start position inside range?
		cp a,c								
		brsh str_find_chr_not_found
str_find_chr_start:
		ldi zl,LOW(STR_BUF)					; Set pointer to start position
		ldi zh,HIGH(STR_BUF)
		add zl,a							
		adc zh,zero			
str_find_chr_loop:
		ld temp,z+							; Has been found?
		cp temp,b
		breq str_find_chr_exit
		inc a
		cpse a,c
		rjmp str_find_chr_loop
str_find_chr_not_found:
		ser a
str_find_chr_exit:
		pop zh
		pop zl
		pop c
		pop b
		pop temp
		ret
#endif

; ### Trims End of String using given Character
; >a: The character to use for trimming String
#ifndef STR_FN_TRIM
#define STR_FN_TRIM
str_trim:
		push a
		push b
		push c 
		push zl
		push zh
		
str_trim_loop:
		lds c,STR_LEN						; Check for empty String?
		tst c
		breq str_trim_loop_exit
		ldi zl,LOW(STR_BUF)					; Is last Char same as given one?
		ldi zh,HIGH(STR_BUF)
		add zl,c
		adc zh,zero
		ld b,-z
		cp a,b
		brne str_trim_loop_exit
		dec c								; Yes, decrement Length
		sts STR_LEN,c
		rjmp str_trim_loop

str_trim_loop_exit:
		pop zh
		pop zl
		pop c
		pop b
		pop a
		ret
#endif

; ### Shrinks String using given Character
; >a: The character to use for shrinking String
#ifndef STR_FN_SHRINK
#define STR_FN_SHRINK
str_shrink:
		push a
		push b
		push c
		push d
		push e
		push f
		push yl
		push yh
		push zl
		push zh

		lds c,STR_LEN						; Cancel, if String is empty!
		tst c			
		breq str_shrink_exit

		mov b,a	
		clr d								; Reset flag for 'Double Trim Chars'
		clr e								; Reset current position inside String
		ldi zl,LOW(STR_BUF)
		ldi zh,HIGH(STR_BUF)
str_shrink_loop:		
		cp e,c								; Is position inside String?
		brsh str_shrink_done
		ld a,z
		cp a,b								; Should Trim?
		brne str_shrink_not_found
		tst d
		breq str_shrink_found
str_shrink_delete:							; Delete 'Double Trim Chars'
		mov yl,zl						
		mov yh,zh
		mov f,e
str_shrink_delete_loop:
		ldd a,y+1
		st y+,a
		inc f
		cp f,c
		brlo str_shrink_delete_loop
str_shrink_delete_done:
		dec c
		ld a,z
		cp a,b
		brne str_shrink_not_found
		cp e,c
		brsh str_shrink_not_found
		rjmp str_shrink_delete
str_shrink_not_found:							
		clr d
		rjmp str_shrink_next
str_shrink_found:
		ser d
str_shrink_next:								; Do next char inside String
		adiw z,1
		inc e
		rjmp str_shrink_loop
str_shrink_done:
		sts STR_LEN,c							; Saves back new String Length
		mov a,b									; Trim char at the end of the String
		rcall str_trim

str_shrink_exit:
		pop zh
		pop zl
		pop yh
		pop yl
		pop f
		pop e
		pop d
		pop c
		pop b
		pop a
		ret
#endif

; *****************************************************************************
; *** Private functions                                                     ***
; *****************************************************************************

#endif ; __STR_INC
