#ifndef __UART_INC
#define __UART_INC

.include "util.inc"
.include "str.inc"

; #############################################################################
; ###                                                                       ###
; ###                         UART (Serial Port)                            ###  
; ###                                                                       ###
; #############################################################################

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************

.ifndef UART_BAUD							; Baudrate
.equ UART_BAUD = 115200						
.endif

; Max. +/-0,1% Error
.equ UART_UBRR_VAL = ((FC+UART_BAUD*8)/(UART_BAUD*16)-1) 
.equ UART_BAUD_REAL = (FC/(16*(UART_UBRR_VAL+1)))     
.equ UART_BAUD_ERROR = ((UART_BAUD_REAL*1000)/UART_BAUD-1000)  
.if ((UART_BAUD_ERROR>10) || (UART_BAUD_ERROR<-10))	
  .error "Baudrate error!"
.endif

; ### FIX SYMBOLS
.ifndef UCSR0A
	.equ UMSEL00 = UMSEL
	.equ UMSEL01 = UMSEL
	.equ UPM00 = UPM0
	.equ UPM01 = UPM1
	.equ USBS0 = USBS
	.equ UCSZ00 = UCSZ0
	.equ UCSZ01 = UCSZ1
	.equ UCSZ02 = UCSZ2
	.equ UCPOL0 = UCPOL
	.equ UCSR0C = UCSRC
	.equ RXCIE0 = RXCIE
	.equ TXCIE0 = TXCIE
	.equ UDRIE0 = UDRIE
	.equ RXEN0 = RXEN
	.equ TXEN0 = TXEN
	.equ RXB80 = RXB8
	.equ TXB80 = TXB8
	.equ UDRE0 = UDRE
.endif

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

.cseg
; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

; *****************************************************************************
; *** Public functions                                                      ***
; *****************************************************************************

; ### Inits UART
#ifndef UART_FN_INIT
#define UART_FN_INIT
uart_init:
		push a

		sbrc init0,INIT0_UART				; Check, if initialization has been already done
		rjmp uart_init_done
		
		mcall util_init						; Init depending modules
		mcall str_init

		ldi a,HIGH(UART_UBRR_VAL)			; Set baudrate from 16bit value of UBRR_VAL
.ifdef UBRR0H
		outreg UBRR0H,a
.else
		outreg UBRRH,a
.endif
		ldi a,LOW(UART_UBRR_VAL)
.ifdef UBRR0L
		outreg UBRR0L,a
.else
		outreg UBRRL,a
.endif
		
		clr a
.ifdef URSEL
		sbr a,(1<<URSEL)
.endif
		sbr a,(0<<UMSEL01)|(0<<UMSEL00)		; UMSEL (USART mode select)     -> 0: Asynchronous USART / 1: Synchronous USART 
		sbr a,(0<<UPM01)|(0<<UPM00)			; UPM1:0 (Parity mode)          -> 00: Disabled / 10: Enabled, even parity / 11: Enabled, odd parity
		sbr a,(0<<USBS0)					; USBS (Stop bit select)        -> 0: 1-bit / 1: 2-bit
		sbr a,(1<<UCSZ01)|(1<<UCSZ00)		; UCSZ1:0 (Character size)      -> 00: 5bit / 01: 6bit / 10: 7bit / 11: 8bit
		sbr a,(0<<UCPOL0)					; UCPOL (Clock polarity)        -> 0: TxD>Rising XCK edge & RxD>Falling XCK edge / 1: TxD>Falling XCK edge & RxD>Rising XCK edge
.ifdef UCSR0C
		outreg UCSR0C,a
.else
		outreg UCSRC,a
.endif
 		clr a
#ifndef UART_NO_IRQ
		sbr a,(1<<RXCIE0)					; RXCIE (RX complete interrupt enable)
#else
		sbr a,(0<<RXCIE0)					; RXCIE (RX complete interrupt disable)
#endif
		sbr a,(0<<TXCIE0)					; TXCIE (TX complete interrupt enable)
		sbr a,(0<<UDRIE0)					; UDRIE (USART data register empty interrupt enable)
		sbr a,(1<<RXEN0)					; RXEN (Receiver enable)
		sbr a,(1<<TXEN0)					; TXEN (Transmitter enable)
		sbr a,(0<<UCSZ02)					; UCSZ2 (Character size for 9bit)
		sbr a,(0<<RXB80)					; RXB8 (Receive data bit 8)
		sbr a,(0<<TXB80)					; TXB8 (Transmit data bit 8)
.ifdef UCSR0B
		outreg UCSR0B,a
.else
		outreg UCSRB,a
.endif

uart_init_done:
		mov a,init0							; Mark as initialized
		sbr a,(1<<INIT0_UART)
		mov init0,a

		pop a
		ret
#endif

#ifdef UART_NO_IRQ
#ifndef UART_FN_GET
#define UART_FN_GET
; ### Get Byte 
; >a: byte to send
uart_get:
		push b

		clr a
.ifdef UCSR0A
		inreg b,UCSR0A
		sbrs b,RXC0
		rjmp uart_get_exit
		inreg a,UDR0							
.else
		inreg b,UCSRA
		sbrs b,RXC
		rjmp uart_get_exit
		inreg a,UDR							
.endif
uart_get_exit:
		pop b
		ret
#endif
#endif

; ### Send Byte 
; >a: byte to send
#ifndef UART_FN_SEND
#define UART_FN_SEND
uart_send:
		push b

uart_send_loop:
.ifdef UCSR0A
		inreg b,UCSR0A
		sbrs b,UDRE0
  		rjmp uart_send_loop
  		outreg UDR0,a 
.else
		inreg b,UCSRA
		sbrs b,UDRE
  		rjmp uart_send_loop
  		outreg UDR,a 
.endif

		pop b
		ret
#endif

; ### Send New Line
#ifndef UART_FN_SEND_NL
#define UART_FN_SEND_NL
uart_send_nl:
		push a
		
		ldi a,CHR_CR					
		rcall uart_send
		ldi a,CHR_LF					
		rcall uart_send

		pop a
		ret
#endif

; ### Send String Buffer
#ifndef UART_FN_SEND_STR
#define UART_FN_SEND_STR
uart_send_str:
		push a
		push b
		push c

		mcall str_get_len
		mov c,a
		tst c
		breq uart_send_str_exit
		clr b
uart_send_str_loop:
		mcall str_get
		cpse a,zero
		rcall uart_send
		inc b
		cp b,c
		brlo uart_send_str_loop

uart_send_str_exit:
		pop c
		pop b
		pop a
		ret
#endif

; *****************************************************************************
; *** Private functions                                                     ***
; *****************************************************************************

#endif ; __UART_INC
