#ifndef __UTIL_INC
#define __UTIL_INC

; ### Check for FC symbol
.ifndef FC
.error "FC with CPU frequency in Hz is not defined!"
.endif

; #############################################################################
; ###                                                                       ###
; ###                              Utility                                  ###  
; ###                                                                       ###
; #############################################################################

; !!! Use define UTIL_DEC_IGNORE for ignoring some decimal functions !!!

; *****************************************************************************
; *** Definitions                                                           ***
; *****************************************************************************

.def temp = r0								; temporary register (memory load)
.def zero = r1								; always zero
.def stat = r2								; temporary register (status register)
.def init0 = r3								; init register #0 0(Core Init)
.def a0 = r4								; 32bit accu register (byte #0) LSB
.def a1 = r5								; 32bit accu register (byte #1)
.def a2 = r6								; 32bit accu register (byte #2)
.def a3 = r7								; 32bit accu register (byte #3) MSB
.def o0 = r8								; 32bit operand register (byte #0) LSB
.def o1 = r9								; 32bit operand register (byte #1)
.def o2 = r10								; 32bit operand register (byte #2)
.def o3 = r11								; 32bit operand register (byte #3) MSB
.def t0 = r12								; 32bit temporary register (byte #0) LSB
.def t1 = r13								; 32bit temporary register (byte #1)
.def t2 = r14								; 32bit temporary register (byte #2)
.def t3 = r15								; 32bit temporary register (byte #3) MSB
.def a = r16								; temporary register (a)	
.def b = r17								; temporary register (b)
.def c = r18								; temporary register (c)
.def d = r19								; temporary register (d)
.def e = r20								; temporary register (e)
.def f = r21								; temporary register (f)
.def g = r22								; temporary register (g)
.def pdat = r23								; temporary port register
.def ol = r24								; counter register (LO)
.def oh = r25								; counter register (HI)

; === Init Bits ===============================================================
.equ INIT0_UTIL = 0							; INIT0 Bit 0: Utility Module
.equ INIT0_STR = 1							; INIT0 Bit 1: String Module
.equ INIT0_MATH = 2							; INIT0 Bit 2: Math Module
.equ INIT0_CLOCK = 3						; INIT0 Bit 3: Clock Module
.equ INIT0_UART = 4							; INIT0 Bit 4: UART Module
.equ INIT0_SOFTUART = 5						; INIT0 Bit 5: SOFT UART Module
.equ INIT0_I2C = 6							; INIT0 Bit 6: I2C Module
.equ INIT0_APPLICATION = 7					; INIT0 Bit 7: Application Module (Top-level)

.equ INIT1_MCP23S17 = 0						; INIT1 Bit 0: MCP23S17 (Port Expander via SPI)
.equ INIT1_IDE = 1							; INIT1 Bit 1: IDE (Interface with Data via MCP23S17)
.equ INIT1_EE24C08 = 2						; INIT1 Bit 2: EE24C08 Module (EEPROM 24C08 / 1 KB)
.equ INIT1_DS18B20 = 3						; INIT1 Bit 3: DS18B20 Module (Temperature Sensor)
.equ INIT1_TNARX = 4						; INIT1 Bit 4: TNARX Module (LCD UC121902-TNARX-A / 12 Digits + Symbols)
.equ INIT1_HD44780 = 5   					; INIT1 BUt 5: HD44780 Module (LCD HD44780 / From 8x1 to 40x4 Character Display) 
.equ INIT1_NRF24L01 = 6						; INIT1 Bit 6: NRF24L01 Module (RF NRF24L01+ / 2,4 GHz)
.equ INIT1_RFM12B = 7						; INIT1 Bit 7: RFM12B Module (RFM12B / 433 MHz & 868 MHz)

.equ INIT2_DHT21 = 0						; INIT2 Bit 0: DHT21 Module (Temperature & Humidity Sensor via 1-wire)
.equ INIT2_SED1520 = 1						; INIT2 Bit 1: SED1520 Module (LCD Controller for 122 x 32 Pixel Display)
.equ INIT2_SAA1064 = 2						; INIT2 Bit 2: SAA1064 Module (LED Display Driver 16Bit)
.equ INIT2_MCP7940M = 3						; INIT2 Bit 3: MCP7940M Module (Realtime Clock with 56 Bytes SRAM)
.equ INIT2_DATE = 4							; INIT2 Bit 4: DATE Module (Handling Dates)
.equ INIT2_EE24C64 = 5						; INIT2 Bit 5: EE24C64 Module (EEPROM 24C64 / 8 KB)
.equ INIT2_CHIP8 = 6						; INIT2 Bit 6: CHIP8 Modue  (VM for Chip-8 Code)
.equ INIT2_RF433 = 7						; INIT2 Bit 7: RF433 (RF 433MHZ Transmitter)

.equ INIT3_BMP085 = 0						; INIT3 Bit 0: BMP085 Module (BMP085/BMP180 Pressure Sensor)
.equ INIT3_CRC8 = 1							; INIT3 Bit 1: CRC8 Module (Dallas 1-Wire CRC)

; === Control Characters (ASCII) ==============================================
.equ CHR_FF = 0x0C							; Form Feed (Clear Screen)
.equ CHR_CR = '\r'							; Carrige Return
.equ CHR_LF = '\n'							; Line Feed
.equ CHR_ESC = 0x1B							; Escapce

; *****************************************************************************
; *** Macros                                                                ***
; *****************************************************************************

; ### Adds Byte (1) to Register (0)
.macro adi 
	subi @0, -(@1)
.endm

; ### Adds Word (1) to Register Pair (0H:0L)
.macro addiw
    subi @0L, LOW(-@1)  
    sbci @0H, HIGH(-@1)
.endm

; ### Perform call (0) to other module
.macro mcall 
.if (FLASHEND > 0x0fff)
   call @0
.else
   rcall @0
.endif
.endm

; ### Loads Long into register for MATH module
.macro ldil 
	ldi a,byte1(@0)
	ldi b,byte2(@0)
	ldi c,byte3(@0)
	ldi d,byte4(@0)
.endm

; ### Reads input (0) with register (1)
.macro inreg
    .if @1 < 0x40
        in @0, @1
    .elif ((@1 >= 0x60) && (@1 < SRAM_START))
        lds @0,@1
    .else
       .error "Invalid I/O register address"
    .endif
.endmacro

; ### Writes output (1) to register (0)
.macro outreg
    .if @0 < 0x40
        out @0, @1
    .elif ((@0 >= 0x60) && (@0 < SRAM_START))
        sts @0,@1
    .else
       .error "Invalid I/O register address"
    .endif
.endmacro 

.dseg              
; *****************************************************************************
; *** SRAM                                                                  ***
; *****************************************************************************

UTIL_INIT1:			.byte 1					; Initialization Register #1
UTIL_INIT2:			.byte 1					; Initialization Register #2
UTIL_INIT3:			.byte 1					; Initialization Register #3
UTIL_TMP_DIGITS:	.byte 5					; Temporary Buffer for holding Digits of 16bit decimal (5 Digits)
UTIL_RANDOM_NUM:	.byte 1					; Last Random Number

.cseg
; *****************************************************************************
; *** DATA                                                                  ***
; *****************************************************************************

UTIL_CENTS:									; Hundreds
	.dw 0,100,200,300,400,500,600,700,800,900
UTIL_TENTHS:								; Tenths
	.db 0,10,20,30,40,50,60,70,80,90
UTIL_DIGITS:								; Digit Characters (decimal and hex)
	.db "0123456789ABCDEF"

; *****************************************************************************
; *** Public functions                                                      ***
; *****************************************************************************

; ### Inits utility
#ifndef UTIL_FN_INIT
#define UTIL_FN_INIT
util_init:
		push a
		
		sbrc init0,INIT0_UTIL				; Check, if initialization has been already done
		rjmp util_init_done
		
		clr zero							; Reset zero register
		sts UTIL_INIT1,zero					; Reset init register 
		sts UTIL_INIT2,zero	
		sts UTIL_INIT3,zero
					
		; TODO

util_init_done:
		mov a,init0							; Mark as initialized
		sbr a,(1<<INIT0_UTIL)
		mov init0,a

		pop a
		ret
#endif

; ### Delay 10us
#ifndef UTIL_FN_DELAY_10US
#define UTIL_FN_DELAY_10US
util_delay_10us:
.if FC>=1000000
; >>> more than 1MHz
	.if FC>=2000000
	.equ CLK10US = ((FC/100000)-12)/3		; Constant for 10us including RCALL
	.else
	.equ CLK10US = 1							
	.endif
			push a							; 1 x push = 2 clocks

			ldi a,CLK10US					; 1 x load = 1 clock
util_delay_10us_loop:						
			dec a							; N * 3 clocks
			brne util_delay_10us_loop

			pop a							; 1 x pop + ret = 6 clocks
			ret
.else
; >>> lower than 1MHz
; TODO calculating NOPs lower than 1Mhz?
			ret			
.endif
#endif

; ### Delay X*10us
; (tolerance by additional 2+N*(0+1+2)+2+4
; >a: value of ms
#ifndef UTIL_FN_DELAY_US
#define UTIL_FN_DELAY_US
util_delay_X10us:
		push a								 

util_delay_X10us_loop:						; Delays X times for 10 us
		rcall util_delay_10us				; Cycles for RCALL are already included in calculation here
		dec a
		brne util_delay_X10us_loop

		pop a
		ret
#endif

; ### Delay 1ms
#ifndef UTIL_FN_DELAY_1MS
#define UTIL_FN_DELAY_1MS
.equ CLK1MS = ((FC/1000)-17)/4				; Constant for 1ms including RCALL
util_delay_1ms: 
		push ol								; 2 x push = 4 clocks
		push oh								

		ldi oh,HIGH(CLK1MS)					; 2 x load = 2 clocks
		ldi ol,LOW(CLK1MS)
util_delay_1ms_loop:
		sbiw ol,1							; N * 4 clocks
		brne util_delay_1ms_loop 

		pop oh								; 2 x pop + ret = 8 clocks 
		pop ol
		ret
#endif

; ### Delay 10ms
; (tolerance by additional 2+1+0+2+4 = 9 cycles)
#ifndef UTIL_FN_DELAY_10MS
#define UTIL_FN_DELAY_10MS
util_delay_10ms:
		push a							

		ldi a,10						
		rcall util_delay_Xms				; Cycles for RCALL are already included in calculation here

		pop a							
		ret
#endif

; ### Delay X*0-255 ms
; (tolerance by additional 2+X*(0+1+2)+2+4
; >a: value of ms
#ifndef UTIL_FN_DELAY_MS
#define UTIL_FN_DELAY_MS
util_delay_Xms:
		push a

util_delay_Xms_loop:						; Delays X times for one millisecond
		rcall util_delay_1ms				; Cycles for RCALL are already included in calculation here
		dec a
		brne util_delay_Xms_loop

		pop a
		ret
#endif

; ### Read Byte from PROM to SRAM
; <a: byte read
; >z: Pointer inside PROM
#ifndef UTIL_FN_GET_EEPROM
#define UTIL_FN_GET_EEPROM
util_get_eeprom:
.ifdef EEPE
		sbic EECR,EEPE						; Check for Previous Write Access          
		rjmp util_get_eeprom
.else
		sbic EECR,EEWE
		rjmp util_get_eeprom
.endif		
.ifdef EEARH        
		out EEARH,zh
.endif
		out EEARL,zl
		sbi EECR,EERE						; Activate Read Access
		in a,EEDR				 
		ret
#endif

; ### Converts Hexadecimal to Byte (2 digits)
; <a: Byte (0-255)
; >b: Digit #1
; >c: Digit #2
#ifndef UTIL_FN_HEX2BYTE
#define UTIL_FN_HEX2BYTE
util_hex2byte:
		push d
		push temp
		push zl
		push zh

		clr a
util_hex2byte1:
		clr d								; Digit #1
		ldi zl,LOW(2*UTIL_DIGITS)				
		ldi zh,HIGH(2*UTIL_DIGITS)
util_hex2byte1_loop:
		lpm
		adiw z,1
		inc d
		cpi d,0x11
		breq util_hex2byte2
		cp temp,b
		brne util_hex2byte1_loop
		dec d
		add a,d
		swap a
util_hex2byte2:
		clr d								; Digit #2
		ldi zl,LOW(2*UTIL_DIGITS)				
		ldi zh,HIGH(2*UTIL_DIGITS)
util_hex2byte2_loop:
		lpm
		adiw z,1
		inc d
		cpi d,0x11
		breq util_hex2byte_exit
		cp temp,c
		brne util_hex2byte2_loop
		dec d
		add a,d

util_hex2byte_exit:
		pop zh
		pop zl
		pop temp
		pop d
		ret	   
#endif

; ### Converts Byte to Hexadecimal (2 digits)
; >a: Byte
; <b: Digit #1 
; <c: Digit #2
#ifndef UTIL_FN_BYTE2HEX
#define UTIL_FN_BYTE2HEX
util_byte2hex:
		push temp
		push zl
		push zh

		ldi zl,LOW(2*UTIL_DIGITS)			; Digit #1
		ldi zh,HIGH(2*UTIL_DIGITS)
		mov b,a
		swap b
		andi b,0x0F
		add zl,b
		adc zh,zero
		lpm
		mov b,temp

		ldi zl,LOW(2*UTIL_DIGITS)			; Digit #2
		ldi zh,HIGH(2*UTIL_DIGITS)
		mov c,a
		andi c,0x0F
		add zl,c
		adc zh,zero
		lpm
		mov c,temp

		pop zh
		pop zl
		pop temp
		ret	
#endif

; ### Converts Byte to Nibble in Hexadecimal (1 digit)
; <>a: Byte (Nibble) as Input / Output is Nibble as ASCII Digit
#ifndef UTIL_FN_BYTE2NIBBLE
#define UTIL_FN_BYTE2NIBBLE
util_byte2nibble:
		push temp
		push zl
		push zh

		ldi zl,LOW(2*UTIL_DIGITS)			; Digit #1
		ldi zh,HIGH(2*UTIL_DIGITS)
		andi a,0x0F
		add zl,a
		adc zh,zero
		lpm
		mov a,temp

		pop zh
		pop zl
		pop temp
		ret	
#endif 
 

; ### Converts Byte to Decimal (3 digits)
; >a: The Byte to convert
; <b: The Digit #1
; <c: The Digit #2
; <d: The Digit #3
; >>Leading Zeros are replaced by spaces ' '
#ifndef UTIL_FN_BYTE2DEC
#define UTIL_FN_BYTE2DEC
#ifndef UTIL_DEC_IGNORE
util_byte2dec:
		push a
		push ol
		push oh
		
		mov ol,a
		clr oh
		rcall util_word2dec
		lds b,UTIL_TMP_DIGITS+2
		lds c,UTIL_TMP_DIGITS+3
		lds d,UTIL_TMP_DIGITS+4

		pop oh
		pop ol
		pop a	
		ret
#endif
#endif

; ### Converts 16-bit Word to Decimal (5 digits)
; >o: The 16bit-Word to convert
; >>ASCII Digits are stored in UTIL_TMP_DIGITS!
; >>Leading Zeros are replaced by spaces ' '
#ifndef UTIL_FN_WORD2DEC
#define UTIL_FN_WORD2DEC
#ifndef UTIL_DEC_IGNORE
util_word2dec:
		push a
		push b
		push zl
		push zh

		rcall util_word2bcd					; Convert 16bit-Word to BCD
		
		ldi zl,LOW(UTIL_TMP_DIGITS)
		ldi zh,HIGH(UTIL_TMP_DIGITS)
		ldi b,4
util_word2dec_lead:
		ld a,z								; Check for leading zero BCD Digit
		tst a 
		brne util_word2dec_cont 
		ldi a,' '							; Mark leading zero BCD Digit by space ' '
		st z+,a
		dec b
		brne util_word2dec_lead
		ld a,z								; Get current BCD Digit
util_word2dec_cont:
		inc b
util_word2dec_loop:
		adi a,'0'							; Add ASCII '0'
		st z+,a
		ld a,z 
		dec b
		brne util_word2dec_loop

		pop zh
		pop zl
		pop b
		pop a
		ret
#endif
#endif

; ### Converts Byte to BCD (3 digits)
; >a: The Byte to convert
; <b: The BCD Digit #1
; <c: The BCD Digit #2
; <d: The BCD Digit #3
#ifndef UTIL_FN_BYTE2BCD
#define UTIL_FN_BYTE2BCD
#ifndef UTIL_DEC_IGNORE
util_byte2bcd:
		push a
		push ol
		push oh
		
		mov ol,a
		clr oh
		rcall util_word2bcd
		lds b,UTIL_TMP_DIGITS+2
		lds c,UTIL_TMP_DIGITS+3
		lds d,UTIL_TMP_DIGITS+4

		pop oh
		pop ol
		pop a	
		ret
#endif
#endif

; ### Converts Byte to BCD (2 digits)
; <>a: The Byte
#ifndef UTIL_FN_BYTE2BCD2
#define UTIL_FN_BYTE2BCD2
#ifndef UTIL_DEC_IGNORE
util_byte2bcd2:
		push ol
		push oh
		
		mov ol,a
		clr oh
		rcall util_word2bcd

		lds a,UTIL_TMP_DIGITS+3
		swap a
		lds ol,UTIL_TMP_DIGITS+4
		or a,ol

		pop oh
		pop ol
		ret
#endif
#endif

; ### Converts 16-bit Word to BCD (5 digits)
; >o: The 16bit-Word to convert
; >>BCD Digits are stored in UTIL_TMP_DIGITS!
#ifndef UTIL_FN_WORD2BCD
#define UTIL_FN_WORD2BCD
#ifndef UTIL_DEC_IGNORE
util_word2bcd:
		push ol
		push oh
		push xl
		push xh
		push zl
		push zh

		ldi zl,LOW(UTIL_TMP_DIGITS)
		ldi zh,HIGH(UTIL_TMP_DIGITS)
		ldi xl,LOW(10000)					; Digit #1 (10000)
		ldi xh,HIGH(10000)
		rcall _util_bin2digit
		ldi xl,LOW(1000)					; Digit #2 (1000)
		ldi xh,HIGH(1000)
		rcall _util_bin2digit
		ldi xl,LOW(100)						; Digit #3 (100)
		ldi xh,HIGH(100)
		rcall _util_bin2digit
		ldi xl,LOW(10)						; Digit #4 (10)
		ldi xh,HIGH(10)
		rcall _util_bin2digit
		st z,ol								; Digit #5 (1)

		pop zh
		pop zl
		pop xh
		pop xl
		pop oh
		pop ol
		ret
#endif  
#endif 

; ### Converts Decimal (3 digits) to Byte
; <<carry: If carry-flag is set, there is an invalid character
; <a: Byte
; >b: Digit #1 (100)
; >c: Digit #2 (10)
; >d: Digit #3 (1)
#ifndef UTIL_FN_DEC2BYTE
#define UTIL_FN_DEC2BYTE
#ifndef UTIL_DEC_IGNORE
util_dec2byte:
		push temp
		push b
		push c
		push d
		push zl
		push zh

		cpi b,'0'							; Check Digit #1 (100)
		brlo util_dec2byte_invalid
		cpi b,'3'		
		brsh util_dec2byte_invalid
		subi b,'0'
		ldi zl,LOW(2*UTIL_CENTS)		
		ldi zh,HIGH(2*UTIL_CENTS)
		lsl b
		add zl,b
		adc zh,zero
		lpm
		mov a,temp
		
		cpi c,'0'							; Check Digit #2 (10)
		brlo util_dec2byte_invalid
		cpi c,'0'+10
		brsh util_dec2byte_invalid
		subi c,'0'
		ldi zl,LOW(2*UTIL_TENTHS)		
		ldi zh,HIGH(2*UTIL_TENTHS)
		add zl,c
		adc zh,zero
		lpm
		add a,temp
		brcs util_dec2byte_invalid

		cpi d,'0'							; Check Digit #3 (1)
		brlo util_dec2byte_invalid
		cpi d,'0'+10
		brsh util_dec2byte_invalid
		subi d,'0'
		add a,d
		brcs util_dec2byte_invalid
		rjmp util_dec2byte_exit

util_dec2byte_invalid:
		clr a
		sec 
util_dec2byte_exit:
		pop zh
		pop zl
		pop d
		pop c
		pop b
		pop temp
		ret
#endif
#endif

; ### Converts BCD (2 digits) to Byte
; <<carry: If carry-flag is set, there is an invalid character
; <>a: Byte (Both Nibbles)
#ifndef UTIL_FN_BCD2BYTE
#define UTIL_FN_BCD2BYTE
#ifndef UTIL_DEC_IGNORE
util_bcd2byte:
		push temp
		push b
		push c
		push zl
		push zh

		mov b,a								; Extract 2 Digits by Nibbles
		swap b
		andi b,0x0F
		mov c,a
		andi c,0x0F
										
		cpi b,10							; Check Digit #1 (10)
		brsh util_bcd2byte_invalid
		ldi zl,LOW(2*UTIL_TENTHS)		
		ldi zh,HIGH(2*UTIL_TENTHS)
		add zl,b
		adc zh,zero
		lpm
		mov a,temp
		brcs util_bcd2byte_invalid

		cpi c,10							; Check Digit #2 (1)
		brsh util_bcd2byte_invalid
		add a,c
		brcs util_bcd2byte_invalid
		rjmp util_bcd2byte_exit

util_bcd2byte_invalid:
		clr a
		sec 
util_bcd2byte_exit:
		pop zh
		pop zl
		pop c
		pop b
		pop temp
		ret
#endif
#endif

; ### Randomize
; >a: Byte 
#ifndef UTIL_FN_RANDOMIZE
#define UTIL_FN_RANDOMIZE
util_randomize:
		sts UTIL_RANDOM_NUM,a
		ret
#endif

; ### Gets Random Number
; <a: Byte 
#ifndef UTIL_FN_RAND
#define UTIL_FN_RAND
util_rand:
		push b

		lds a,UTIL_RANDOM_NUM
		rol a
		brcc util_rand_exit 
		ldi b,29 
		eor a,b 
util_rand_exit:
		sts UTIL_RANDOM_NUM,a

		pop b
		ret
#endif

; *****************************************************************************
; *** Private functions                                                     ***
; *****************************************************************************

; ### Converts 16bit-word for current BCD digit
; >x: The 16bit decimal for current Digit
; <>o: The current 16-bit word as value (rest by division of o)
; <>z: Pointer to resulting BCD Digit inside temporary buffer
#ifndef UTIL_FN__BIN2DIGIT  
#define UTIL_FN__BIN2DIGIT
#ifndef UTIL_DEC_IGNORE
_util_bin2digit:
		push a

		clr a		
_util_bin2digit_loop:
		cp oh,xh							; Compare MSBs
		brcs _util_bin2digit_exit
		brne _util_bin2digit_loop_sub
		cp ol,xl							; MSBs are equal, compare LSBs
		brcs _util_bin2digit_exit
_util_bin2digit_loop_sub:
		sub ol,xl 
		sbc oh,xh
		inc a
		rjmp _util_bin2digit_loop
_util_bin2digit_exit:
		st z+,a								; Save resulting BCD Digit to temporary buffer

		pop a
		ret
#endif
#endif

#endif ; __UTIL_INC
